#include "alg.h"

////////////////////////////////////////////////////////////////////////////////

void Algorithm::RunAlg(ALG_TYPE alg_type)
{
#ifdef FUNCTION_TRACE
  fprintf(stderr, "++ RunAlg() \n");
#endif

  switch(alg_type) 
  {
    case PROPOSED_ALG:
    case SHORTEST_PATH:
      StartOurAlgorithm();
      break;
    default:
      fprintf(stderr, "Unimplemented algorithm !\n");
      break;
  }
}  // end RunAlg()


////////////////////////////////////////////////////////////////////////////////
//
// Method: StartOurAlgorithm
//
// Description:
// for every request group
//   for every request in this group
//   { check validity of session ID, requesting node ID, request type, requested QoS;
//     if request type is JOIN
//       call StartJoinAlgorithm();
//     else if request type is LEAVE
//       call StartLeaveAlgorithm();
//     else "Invalid request type";
//   }
//
// Note: A future implementation
//	Since the request file can be very big, we will read and process 
//	request groups one at a time to save memory.
//
////////////////////////////////////////////////////////////////////////////////

void Algorithm::StartOurAlgorithm(void)
{
#ifdef FUNCTION_TRACE
  fprintf(stderr, "++++ StartOurAlgorithm() \n");
#endif

  GrpReq*  grp_req;
  UserReq* userRequest;
  Session* session;
  BOOL     retBool;

  printf("\n[[ Start processing all request groups ]]\n");

  // For every request group, process the group
  for(grp_req  = data_mgr->GetGrpReqHead();
      grp_req != NIL;
      grp_req  = grp_req->GetNext())  		 // grp_req = grp_req->next;
  {
    // Sort request in this group if required
    // retBool = grp_req->SortRequests();

    // For every request in this group, process the request
    for(userRequest  = grp_req->GetUserReqHead();
	userRequest != NIL;
	userRequest  = userRequest->GetNext())   // userRequest = userRequest->next;
    {
      // Debugging traces
      printf("\n=============================================================================\n");
      //printf("PLOT ");
      userRequest->Display();

#ifdef FUNCTION_TRACE
      fprintf(stderr, "\n");
      userRequest->Display(stderr);
#endif

      // Check validity of data in the request
      retBool = CheckRequestValidity(userRequest);
      if (retBool == FALSE) continue;    // ignore this request, get the next one

      // Given session ID, find pointer to Session object
      int req_session_id = userRequest->GetSessionID();  
      session = data_mgr->FindSession(req_session_id);

      // Process the request
      switch(userRequest->GetUserReq())  // consider request type (join, leave)
      {
	case JOIN:
	  retBool = StartJoinAlgorithm(session, userRequest);
	  break;
	case LEAVE:
	  retBool = StartLeaveAlgorithm(session, userRequest);
	  break;
	default:
	  retBool = FALSE;
	  printf("Invalid request type => ignored!\n");
	  break;
      }

      // Display statistics
      if (retBool != FALSE)	// see Note below
      {
	session->UpdateStatistics(data_mgr->GetGraph(), userRequest, retBool);
	session->GetSessionStat()->Display();
      }

#ifdef DEBUG_LEVEL_2
      // Display the MC tree and the paths from mcSource to every destination
      if (retBool == TRUE)
      {
      	session->Display_tree();
      	session->DisplayTreePaths();
      }
#endif

    }  // end for(userRequest)
  }  // end for(grp_req)

  // Display final statistics
  session->GetSessionStat()->DisplayFinalStat();
  return;
}  // end StartOurAlgorithm()


//
// Note:
// - request is ignored    => return code = FALSE
// - request is rejected   => return code = FALSE_2
// - request is successful => return code = TRUE
//

////////////////////////////////////////////////////////////////////////////////

BOOL Algorithm::CheckRequestValidity(UserReq* userRequest)
{
#ifdef FUNCTION_TRACE
  fprintf(stderr, "++++++ CheckRequestValidity() \n");
#endif

  // Check validity of session ID
  int req_session_id = userRequest->GetSessionID();  
  Session* session = data_mgr->FindSession(req_session_id);
	// given session ID, find pointer to Session object
  if (session == NIL)
  {
    printf("Requested session[%d] does not exist => ignored!\n", req_session_id);
    return FALSE;
  }

  // Check validity of requesting node ID (0 <= ID < total number of nodes)  
  if (userRequest->GetNodeID() < 0 ||
      userRequest->GetNodeID() >= data_mgr->GetGraph()->GetNodeNum())
  {
    printf("Invalid node ID => ignored!\n");
    return FALSE;
  }

  // Check validity of request type (join, leave, teardown)  
  if (userRequest->GetUserReq() != JOIN && userRequest->GetUserReq() != LEAVE)
  {
    printf("Invalid request type => ignored!\n");
    return FALSE;
  }

  // Check validity of requested QoS  

  return TRUE;
}


////////////////////////////////////////////////////////////////////////////////

BOOL Algorithm::StartJoinAlgorithm(Session* session, UserReq* userRequest)
{
#ifdef FUNCTION_TRACE
  fprintf(stderr, "++++++ StartJoinAlgorithm() \n");
#endif

  BOOL retBool;

  // If requesting node q is mcSource, ignore the request.
  int reqNodeID = userRequest->GetNodeID();
  if (session->GetSrcNode() == reqNodeID)  
  {
    printf("Requesting node [%d] = mcSource node => ignored! \n", reqNodeID);
    return FALSE;
  }

  // If q is already in the MC tree, and a participant node then done; do nothing.
  // (Here we ignore the required QoS of this request, which may be more demanding
  // than that of the previous request generated by the same node). 
  //
  if (session->GetNodeStatus(reqNodeID) == LEAF_NODE)
  {
    printf("Requesting node [%d] is already a participant => ignored! \n", reqNodeID);
    return FALSE;
  }

  // If q is already in the MC tree, and an intermediate node
  if (session->GetNodeStatus(reqNodeID) == INTREE)
  {
    // Find the paths from mcSource to the other nodes in the current MC tree
    Path* allPaths = session->FindPathFromSrcToDest();
	// a new Path object is alloc in this method => should dealloc it later

    // Find the path from mcSource to requesting node
    OnePath* thisPath = allPaths->FindOnePathEndWith(reqNodeID);

    // If the path from mcSource to q satisfies requested QoS, make q a participant.
    OnePath* newPath = NIL;
    if (IsValidQOS(thisPath, newPath, session, userRequest) == TRUE)
    {
      Edge* edge;

      // 1. Update node status
      session->SetNodeStatus(reqNodeID, LEAF_NODE);

      // 2. Update edge status (in the MC tree)
      int prevNode = thisPath->FindPreceedingNode(reqNodeID);
      if (prevNode < 0)
	fprintf(stderr, "Fatal error: Preceeding node not found in StartJoinAlgorithm()!\n");
      else
	edge = session->FindEdge(prevNode, reqNodeID);

      if (edge != NIL)
	edge->SetEdgeStatus(LEAF_NODE);
      else
	fprintf(stderr, "Fatal error: Edge not found in StartJoinAlgorithm()!\n");	

      // 3. No need to add any edge to the MC tree
      // 4. Add this path to list of tree paths
      retBool = AddPathToTreePaths(thisPath, newPath, session, userRequest);

      // 5. Update bandwidth of affected edges
      retBool = UpdateBandwidth(SHARED_PATH, thisPath, session, userRequest);
      if (retBool != TRUE)
      {
        fprintf(stderr, "FILE[%s], LINE[%d] : Fatal error in UpdateBandwidth()!\n",
			__FILE__, __LINE__);
        exit(1);
      }

      delete allPaths;
      printf("Intermediate node becomes participant node.\nSuccessful join.\n");
      return TRUE;
    }  // end if

    // Else reject the request
    printf("Intermediate node [%d], QoS not met => rejected!\n", reqNodeID);
    return FALSE_2;	// rejected request
  }

  // Process the join request
  retBool = ProcessJoinRequest(session, userRequest);
  if (retBool == TRUE)
    printf("Successful join. \n");
  return retBool;
}  // end StartJoinAlgorithm


////////////////////////////////////////////////////////////////////////////////

BOOL Algorithm::ProcessJoinRequest(Session* session, UserReq* userRequest)
{
#ifdef FUNCTION_TRACE
  fprintf(stderr, "++++++++ ProcessJoinRequest() \n");
#endif

  Path* candidatePaths;
  BOOL retBool;
  int reqNodeID = userRequest->GetNodeID();  // new node to join

  // Run Dijkstra's algorithm for WEIGHT and DELAY
  // Input : ID of requesting node q and the network graph
  // Output: list of shortest paths from q to the other nodes in the network
  //
  // Path* allMinCostPaths  = ShortestPath(BACKWARD, reqNodeID, WEIGHT, userRequest);
  Path* allMinCostPaths = ShortestPath(BACKWARD, reqNodeID, CONGESTION, userRequest);  
  Path* allMinDelayPaths = ShortestPath(BACKWARD, reqNodeID, DELAY , userRequest); 

  // Reverse the paths, making them run from the other nodes to q
  //
  allMinCostPaths->ConvertBackToFor();
  allMinDelayPaths->ConvertBackToFor();

  // Find candidate paths from mcSource to q (i.e. paths satisfying requested QoS)
  // Input : allMinCostPaths and allMinDelayPaths
  // Output: list of candidate NEW paths from a set of connection nodes to q
  //         (These paths will be combined with appropriate shared paths
  //	      to form candidate paths from mcSource to q)
  //
  switch(GetAlg())
  {
    case PROPOSED_ALG:
      candidatePaths = FindCandidateNewPaths(allMinCostPaths, allMinDelayPaths,
						session, userRequest);
      break;

    case SHORTEST_PATH:
      candidatePaths = FindCandidateNewPaths_SP(allMinCostPaths, allMinDelayPaths,
						session, userRequest);
      // Note: candidatePaths contains at most one path (min-cost or min-delay).
      break;
  }  // end switch

  if (candidatePaths == NIL)
  {
    printf("No candidate path found for [%d] => rejected! \n", reqNodeID);
    return FALSE_2;
  }

#ifdef DEBUG_LEVEL_1
  printf("Candidate S2D paths: \n");
  DisplayCandidateS2DPaths(candidatePaths, session, userRequest);
#endif

  // Find the best S2D path from the list of candidate paths
  // Input : candidatePaths, list of candidate NEW paths from a set of 
  //	     connection nodes to q
  // Output: best path from mcSource to q ("best" determined by function JoinCost())
  //
  Path* bestS2DPath = FindBestS2DPath(candidatePaths, session, userRequest);

  // Get the shared segment and the new segment of the best S2D path
  OnePath* newPath    = bestS2DPath->GetPath();   // newPath = bestS2DPath->m_path;
  OnePath* sharedPath = newPath->GetNext();       // sharedPath = newPath->next;

  // Add the new join node to the multicast tree (which includes
  // 1. Update node status
  // 2. Update edge status
  // 3. Add new edges to the MC tree)
  //
  retBool = UpdateTreeWithPath(sharedPath, newPath, session);

  // 4. Add the new path to list of tree paths
  retBool = AddPathToTreePaths(sharedPath, newPath, session, userRequest);

  // 5. Update bandwidth of affected edges
  retBool = UpdateBandwidth(SHARED_PATH, sharedPath, session, userRequest);
  if (retBool != TRUE)
  {
    fprintf(stderr, "FILE[%s], LINE[%d] : Fatal error in UpdateBandwidth()!\n",
			__FILE__, __LINE__);
    exit(1);
  }

  retBool = UpdateBandwidth(NEW_PATH, newPath, session, userRequest);
  if (retBool != TRUE)
  {
    fprintf(stderr, "FILE[%s], LINE[%d] : Fatal error in UpdateBandwidth()!\n",
			__FILE__, __LINE__);
    exit(1);
  }

#ifdef DEBUG_LEVEL_1
  printf("Best S2D path: \n");  
  OnePath* tmp = session->GetTreePaths()->FindOnePathEndWith(reqNodeID);
  tmp->Display();
#endif

  // Delete dynamic data structures no longer needed
  delete allMinCostPaths;
  delete allMinDelayPaths;
  delete candidatePaths;
  delete bestS2DPath;

  return TRUE;
}  // end ProcessJoinRequest


////////////////////////////////////////////////////////////////////////////////

Path* Algorithm::FindCandidateNewPaths_SP(Path* allMinCostPaths, Path* allMinDelayPaths,
					Session* session, UserReq* userRequest)
{
#ifdef FUNCTION_TRACE
  fprintf(stderr, "++++++++++ FindCandidateNewPaths_SP() \n");
#endif

  OnePath* minCostPath;
  OnePath* minDelayPath;
  OnePath* sharedPath1 = new OnePath();
  OnePath* newPath1    = new OnePath();
  OnePath* sharedPath2 = new OnePath();
  OnePath* newPath2    = new OnePath();
  OnePath* candidateNewPath = NIL;
  Path*    candidateNewPaths = NIL;

  // Get mcSource node ID
  int mcSource = session->GetSrcNode();

  // Get the min-cost path from mcSource to requesting node q
  minCostPath = allMinCostPaths->FindOnePathStartWith(mcSource);
  if (minCostPath == NIL)	// no path from mcSource to q => graph is disconnected
  {
    printf("Disconnected graph! \n");
    return NIL;		// no candidate path found
  }

  // Check if the minCostPath contains a loop.
  // At the same time, separate the minCostPath into a sharedPath 
  // and a newPath (if a sharedPath exists).
  // Note: As input to IsCycle_SP(), minCostPath, sharedPath, newPath must not be NIL.
  //
  BOOL considerMinDelayPath = FALSE;
  if (session->IsCycle_SP(minCostPath, sharedPath1, newPath1) == FALSE)
  {
    if (sharedPath1->GetPathNode() == NIL)	// sharedPath1 is NIL
    {
      delete sharedPath1;
      sharedPath1 = NIL;
    }
    else	// compute and record the costs of the sharedPath
      sharedPath1->SetCosts(sharedPath1->FindCost(data_mgr->GetGraph(), NEW_PATH));

    // Compute and record the costs of the newPath
    newPath1->SetCosts(newPath1->FindCost(data_mgr->GetGraph(), NEW_PATH));

#ifdef FUNCTION_TRACE
    if (sharedPath1 != NIL) sharedPath1->Display(stderr);
    newPath1->Display(stderr);
#endif

    // Check QoS requirements
    if (IsValidQOS(sharedPath1, newPath1, session, userRequest) == TRUE)
      candidateNewPath = new OnePath(newPath1);
    else  // minCostPath does not meet QoS requirements
      considerMinDelayPath = TRUE;
  }
  else  // minCostPath contains loop
    considerMinDelayPath = TRUE;

  if (considerMinDelayPath == TRUE)  // consider the min-delay path
  {
    minDelayPath = allMinDelayPaths->FindOnePathStartWith(mcSource);

    // Check if the minDelayPath contains a loop.
    // At the same time, separate the minCostPath into a sharedPath 
    // and a newPath (if a sharedPath exists).
    // Note: As input to IsCycle_SP(), minCostPath, sharedPath, newPath must not be NIL.
    //
    if (session->IsCycle_SP(minDelayPath, sharedPath2, newPath2) == FALSE)
    {
      if (sharedPath2->GetPathNode() == NIL)	// sharedPath2 is NIL
      {
        delete sharedPath2;
        sharedPath2 = NIL;
      }
      else	// compute and record the costs of the sharedPath
        sharedPath2->SetCosts(sharedPath2->FindCost(data_mgr->GetGraph(), NEW_PATH));

      // Compute and record the costs of the newPath
      newPath2->SetCosts(newPath2->FindCost(data_mgr->GetGraph(), NEW_PATH));

#ifdef FUNCTION_TRACE
      if (sharedPath2 != NIL) sharedPath2->Display(stderr);
      newPath2->Display(stderr);
#endif

      // Check QoS requirements
      if (IsValidQOS(sharedPath2, newPath2, session, userRequest) == TRUE)
        candidateNewPath = new OnePath(newPath2);
    }
  }  // end if (considerMinDelayPath)

  if (candidateNewPath != NIL) 
  {
    candidateNewPaths = new Path(candidateNewPath->GetSrcNode());
    candidateNewPaths->AddOnePath(candidateNewPath);
  }

  if (sharedPath1 != NIL) delete sharedPath1;
  if (sharedPath2 != NIL) delete sharedPath2;
  delete newPath1;
  delete newPath2;
  return candidateNewPaths;	// contains at most one path (candidateNewPath)
}  // end FindCandidateNewPaths_SP


////////////////////////////////////////////////////////////////////////////////

Path* Algorithm::FindCandidateNewPaths(Path* allMinCostPaths, Path* allMinDelayPaths,
					Session* session, UserReq* userRequest)
{
#ifdef FUNCTION_TRACE
  fprintf(stderr, "++++++++++ FindCandidateNewPaths() \n");
#endif

  if (allMinCostPaths->GetPath() == NIL)
    printf("Disconnected graph! (node to join [%d]) \n", userRequest->GetNodeID());

  OnePath* newMinCostPath = NIL;
  OnePath* newMinDelayPath = NIL;
  OnePath* sharedPath = NIL;
  OnePath* candidateNewPath = NIL;
  Path*    candidateNewPaths = new Path(session->GetSrcNode());  
				// alloc & init a new object

  // Find the paths from mcSource to the other nodes in the current MC tree
  Path* allSharedPaths = session->FindPathFromSrcToDest();
	// a new Path object is alloc in this method => should dealloc it later.

  // For every newMinCostPath, do the following
  for (newMinCostPath  = allMinCostPaths->GetPath();
       newMinCostPath != NIL;
       newMinCostPath  = newMinCostPath->GetNext())
  {
    // Get the start node (a potential connection node) of this newMinCostPath
    int connectionNode = newMinCostPath->GetSrcNode();

    // If the start node of the path is not in connection set, reject this path.
    if (session->InConnectionSet(connectionNode) == FALSE) continue;

    // Find the shared path from mcSource to connectionNode
    sharedPath = allSharedPaths->FindOnePathEndWith(connectionNode);

    // Find the corresponding newMinDelayPath starting from connectionNode
    newMinDelayPath = allMinDelayPaths->FindOnePathStartWith(connectionNode);

    // Return either newMinCostPath or newMinDelayPath (or NIL pointer if
    // neither path is accepted).
    // This method must allocate a new OnePath object, init it to the returned
    // path, and return (pointer to) the new object to the calling method.
    // This way, we can return the interpolated path of min-cost
    // and min-delay paths in a future implementation.
    //
    candidateNewPath = FindCandidateNewPath(sharedPath, newMinCostPath,
			newMinDelayPath, session, userRequest);

    // Add the returned candidateNewPath (if any) to list of candidateNewPaths
    if (candidateNewPath != NIL)
      candidateNewPaths->AddOnePath(candidateNewPath);

  }  // end for(newMinCostPath)

  delete allSharedPaths;
  if (candidateNewPaths->GetPath() == NIL)	// no candidate paths were found
  {
    delete candidateNewPaths;
    candidateNewPaths = NIL;
  }
  return candidateNewPaths;
}  // end FindCandidateNewPaths


////////////////////////////////////////////////////////////////////////////////
// 
// Method: FindCandidateNewPath
//
// Description:
//   Return either minCostPath or minDelayPath (or NIL if neither path is accepted).
//   A future implementation could be to return the interpolated path of
//   minCostPath and minDelayPath.
//
////////////////////////////////////////////////////////////////////////////////

OnePath* Algorithm::FindCandidateNewPath (OnePath* sharedPath, OnePath* minCostPath,
		OnePath* minDelayPath, Session* session, UserReq* userRequest)
{
#ifdef FUNCTION_TRACE
  fprintf(stderr, "++++++++++++ FindCandidateNewPath() (cnode [%d])\n",
			minCostPath->GetSrcNode());
#endif

  OnePath* candidatePath = NIL;

  // Consider minCostPath first
  if (session->IsCycle(minCostPath) == FALSE &&
      IsValidQOS(sharedPath, minCostPath, session, userRequest) == TRUE)
    // Alloc a new OnePath object (candidatePath), copy data of minCostPath
    // to candidatePath, and return (pointer to) the candidatePath.
    //
      candidatePath = new OnePath(minCostPath);

  else  // consider minDelayPath
    if (session->IsCycle(minDelayPath) == FALSE &&
	IsValidQOS(sharedPath, minDelayPath, session, userRequest) == TRUE)
      candidatePath = new OnePath(minDelayPath);

#ifdef DEBUG_LEVEL_1
  if (candidatePath == NIL)   // neither path is accepted
    printf("No accepted path from connection node [%d]!\n", minCostPath->GetSrcNode());
#endif

    return candidatePath;
}  // end FindCandidateNewPath


////////////////////////////////////////////////////////////////////////////////

Path* Algorithm::FindBestS2DPath(Path* candidatePaths, Session* session, 
					UserReq* userRequest)
{
#ifdef FUNCTION_TRACE
  fprintf(stderr, "++++++++++ FindBestS2DPath()\n");
  candidatePaths->Display(stderr);
#endif

  OnePath* bestSharedPath = NIL;
  OnePath* bestNewPath = NIL;  
  float    bestCost = INFINITY;
  float    currentCost;
  OnePath* newPath;
  OnePath* sharedPath;
  int      connectionNode;

  // Find the paths from mcSource to the other nodes in the current MC tree
  Path* allSharedPaths = session->FindPathFromSrcToDest();
	// a new Path object is alloc in this method => should dealloc it later.

  // Among all candidate paths, select the best S2D path.
  for(newPath  = candidatePaths->GetPath(); 
      newPath != NIL;
      newPath  = newPath->GetNext())
  {
    // Get the connection node ID
    connectionNode = newPath->GetSrcNode();

    // Find the shared path from mcSource to connection node
    sharedPath = allSharedPaths->FindOnePathEndWith(connectionNode);

    // Compute the join cost of the combined path (shared path + new path)
    currentCost = JoinCost(sharedPath, newPath, session, userRequest);

    // Select the best cost so far
    if (currentCost < bestCost)
    {
      bestSharedPath = sharedPath;
      bestNewPath    = newPath;
      bestCost       = currentCost;
    }
  }  // end for(newPath)

  // Generate returned path (best S2D path): alloc a new Path object
  // for the best S2D path.  Add bestSharedPath and then bestNewPath
  // to this Path object, and return (pointer to) the object.
  //
  Path* bestS2DPath = new Path(session->GetSrcNode());

  if (bestSharedPath != NIL)
  {
    OnePath* retSharedPath = new OnePath(bestSharedPath);
    bestS2DPath->AddOnePath(retSharedPath);
  }

  OnePath* retNewPath = new OnePath(bestNewPath);
  bestS2DPath->AddOnePath(retNewPath);
  //
  // retNewPath is added to the front of retSharedPath.
  // The returned best S2D path will then consist of two segments 
  // corresponding to the bestNewPath and bestSharedPath.

#ifdef FUNCTION_TRACE
  fprintf(stderr, "++++++++++ best connection node [%d] from FindBestS2DPath()\n", 
		bestNewPath->GetSrcNode());
#endif

  delete allSharedPaths;
  return bestS2DPath;
}  // end FindBestS2DPath

// Note:
// - We allocate retSharedPath since bestSharedPath belongs to 
//   allSharedPaths, which will be deleted at the end.
// - We allocate retNewPath since bestNewPath belongs to 
//   candidatePaths, which will also be deleted later (in ProcessJoinRequest)
// - retNewPath is added to the front of retSharedPath because retSharedPath
//   may be NIL (but retNewPath cannot be).  In that case, 
//   retNewPath->next = retSharedPath = NIL.


////////////////////////////////////////////////////////////////////////////////

float Algorithm::JoinCost(OnePath* sharedPath, OnePath* newPath, 
				Session* session, UserReq* userRequest)
{
#ifdef FUNCTION_TRACE
  fprintf(stderr, "++++++++++++ JoinCost() (cnode [%d])\n", newPath->GetSrcNode());
#endif

/*
  // Join cost is determined by edge weights
  //
  if (sharedPath != NIL && newPath != NIL)
    return ((float)(sharedPath->GetWeight() + newPath->GetWeight()));
  else if (newPath != NIL)
    return ((float)(newPath->GetWeight()));
  else if (sharedPath != NIL)
    return ((float)(sharedPath->GetWeight()));
*/

  float delayCostFact  = data_mgr->GetParameters()->GetDelayCostFactor();
  float sharedCostFact = data_mgr->GetParameters()->GetSharedCostFactor();
  float addedCostFact  = data_mgr->GetParameters()->GetAddedCostFactor();

  float delayCost   = 0.0;
  float sharedCost  = 0.0;
  float addedCost   = 0.0;

  OnePath* joinPath = Combine2Paths(sharedPath, newPath);
  if (joinPath != NIL)  
    delayCost = DelayCost(joinPath, userRequest);
  if (sharedPath != NIL)
    sharedCost = CongCost(SHARED_PATH, sharedPath, session, userRequest);
  if (newPath != NIL)
    addedCost = CongCost(NEW_PATH, newPath, session, userRequest);

#ifdef DEBUG_LEVEL_3
  printf("............ JoinCost() (cnode [%d])\n", newPath->GetSrcNode());
  float joinCost = (delayCostFact  * delayCost) + (sharedCostFact * sharedCost)
			+ (addedCostFact  * addedCost );
  printf(">>>>>>>>>>>>>> delay cost  = %f \n", delayCost);
  printf(">>>>>>>>>>>>>> shared cost = %f \n", sharedCost);
  printf(">>>>>>>>>>>>>> added cost  = %f \n", addedCost);
  printf(">>>>>>>>>>>>>> join cost   = %f \n", joinCost);		
#endif

  return( (delayCostFact  * delayCost ) 
        + (sharedCostFact * sharedCost)
        + (addedCostFact  * addedCost )   );
}  // end JoinCost


////////////////////////////////////////////////////////////////////////////////

float Algorithm::DelayCost(OnePath* thePath, UserReq* userRequest)
{
#ifdef FUNCTION_TRACE
  fprintf(stderr, "++++++++++++++ DelayCost() \n");
#endif

  if (thePath == NIL)
    return 0.0;
  float delayEpsilon   = data_mgr->GetParameters()->GetDelayEpsilon();
  int delayBound = userRequest->GetQos().GetDelay();
  int e2eDelay = thePath->GetDelay();	// get delay sum of the path
  if (e2eDelay <= delayBound)
    return(1 / ((float)(delayBound - e2eDelay) + delayEpsilon));
  return INFINITY;
}  // end DelayCost

////////////////////////////////////////////////////////////////////////////////

float Algorithm::CongCost(PATH_TYPE pathType, OnePath* thePath, 
				Session* session, UserReq* userRequest)
{
#ifdef FUNCTION_TRACE
  fprintf(stderr, "++++++++++++++ CongCost() (path type [%d]) \n", pathType);
#endif

  int   endNode;
  Node* startGrpNode;
  Edge* grpEdge;

  float  congCost  = 0.0;
  int	 adjustedEdgeUsedBand;
  int    addedBand = userRequest->GetQos().GetBandwidth();
  int    sessionID = session->GetSessID();
  Graph* graph 	   = data_mgr->GetGraph();
  int 	 lastNode  = thePath->GetDestNode();	   // last node of the path

  int startNode = thePath->GetSrcNode();  	   // get start node of the path
  PathNode* endPathNode = thePath->GetPathNode();  // endPathNode = node_list;
  while (endPathNode != NIL)
  {
    endNode      = endPathNode->GetNodeID();   	   // endNode = endPathNode->node_id;
    startGrpNode = graph->FindNode(startNode);
    grpEdge      = graph->FindEdge(startGrpNode, endNode);

    if (grpEdge == NIL)	// should never get inside this "if"
    {
      fprintf(stderr, "FILE[%s] : LINE[%d] Fatal error: NIL edge (edge [%d -> %d])!\n", 
			__FILE__, __LINE__, startNode, endNode);
      exit(1);
    }  // end if

    // Accumulate the cost (sharedCost or addedCost)
    switch(pathType)
    {
      case NEW_PATH:	// parameter "session" is not used
        congCost += grpEdge->CongestionInc(grpEdge->GetEdgeUsedBand(), addedBand);
	break;

      case SHARED_PATH:
	float edgeNormFactor = grpEdge->GetEdgeNormFactor();

	int numDescendants = session->FindNode(endNode)->GetEdgeNum(FORWARD);
	if (endNode == lastNode)  // new path will be joined to the lastNode
	  numDescendants++; 	  // + 1 for the new path to join
	if (numDescendants == 0)
	{
	  fprintf(stderr, "Fatal Error in CongCost(): number of descendants = 0! \n");
	  exit(1);
	}

  	int sessionBand = grpEdge->GetSessionUsedBandwidth(sessionID);
	if (sessionBand <= addedBand)	// sessionBand would be updated to addedBand
	  adjustedEdgeUsedBand = grpEdge->GetEdgeUsedBand() - sessionBand;
	else	// sessionBand would be unchanged after the join
	  adjustedEdgeUsedBand = grpEdge->GetEdgeUsedBand() - addedBand;

  	congCost = congCost + ((edgeNormFactor / (float)numDescendants) *
			grpEdge->CongestionInc(adjustedEdgeUsedBand, addedBand));
	break;
    }  // end switch

    // Get the next node on the path
    startNode = endNode;
    endPathNode = endPathNode->GetNext();
  }  // end while

  return congCost;
}

// Note: 
// For sharedCost, we compute CongestionInc(adjustedEdgeUsedBand, addedBand).
// Formula for adjustedEdgeUsedBand:
// adjustedEdgeUsedBand = (used bandwidth of the other sessions) 
//			+ (used bandwidth of this session AFTER the join)
//			- (requested bandwidth)
//

////////////////////////////////////////////////////////////////////////////////

BOOL Algorithm::AddPathToTreePaths(OnePath* sharedPath, OnePath* newPath, 
				Session* session, UserReq* userRequest)
{
#ifdef FUNCTION_TRACE
  fprintf(stderr, "++++++++++ AddPathToTreePaths() \n");
#endif

  OnePath* joinPath = Combine2Paths(sharedPath, newPath);
  if (joinPath == NIL)
  {
    fprintf(stderr, "Fatal error in AddPathToTreePaths(): should not get here!\n");
    return FALSE;
  }

#ifdef DEBUG_LEVEL_1
  if (joinPath->GetSrcNode() != session->GetSrcNode())
  {
    fprintf(stderr, "Fatal error: inconsistent source node in AddPathToTreePaths()!\n");
    return FALSE;
  }

  if (joinPath->GetDestNode() != userRequest->GetNodeID())
  {
    fprintf(stderr, "Fatal error: inconsistent participant node in AddPathToTreePaths()!\n");
    return FALSE;
  }
#endif

  // Save requested bandwidth
  joinPath->SetDestNodeReqBand(userRequest->GetQos().GetBandwidth());

  // Add joinPath to (the front) of list of paths of MC tree
  session->GetTreePaths()->AddOnePath(joinPath);

  // Update numDependants[] of the nodes on the joinPath
  session->UpdateNumDependants(joinPath, 1);	// 1 for joining (-1 for leaving)

  return TRUE;
}  // end AddPathToTreePaths


////////////////////////////////////////////////////////////////////////////////

OnePath* Algorithm::Combine2Paths(OnePath* path1, OnePath* path2)
{
#ifdef FUNCTION_TRACE
  fprintf(stderr, "++++++++++++++++ Combine2Paths() \n");
#endif

  COST retCosts(0, 0, 0);
  OnePath* joinPath;

  if (path1 == NIL && path2 == NIL)
    return NIL;

  if (path1 == NIL)  // (&& path2 != NIL)
    joinPath = new OnePath(path2);
  else  // (path1 != NIL)
  {
    // Alloc a new OnePath object, init it to "path1"
    // (Attribute "src_node" is copied from "path1" to "joinPath")
    // (Attribute "next" is set to NIL)
    //
    joinPath = new OnePath(path1);

    // Add "path2" to the end of "path1"
    // (Attribute "dest_node" of "joinPath" is copied from "path2")
    // If "path2" is NIL, AddPathToEnd() simply returns without modifying "joinPath".
    //
    joinPath->AddPathToEnd(path2);
  }  // end else

  // Compute and update the costs of the joinPath
  retCosts = joinPath->FindCost(data_mgr->GetGraph(), NEW_PATH);
  joinPath->SetCosts(retCosts); 

  return joinPath;
}  // end Combine2Paths

////////////////////////////////////////////////////////////////////////////////

BOOL Algorithm::StartLeaveAlgorithm(Session* session, UserReq* userRequest)
{
#ifdef FUNCTION_TRACE
  fprintf(stderr, "++++++ StartLeaveAlgorithm() \n");
#endif

  int reqNodeID = userRequest->GetNodeID();  // requesting node ID

  // If requesting node q is not in the MC tree, ignore the request.
  if (session->GetNodeStatus(reqNodeID) != LEAF_NODE &&
      session->GetNodeStatus(reqNodeID) != INTREE)
  {
    printf("Requesting node [%d] not in MC tree => ignored!\n", reqNodeID);
    return FALSE;
  }

  // If q is in the MC tree, but not a participant, ignore the request.
  if (session->GetNodeStatus(reqNodeID) == INTREE)
  {
    printf("Requesting node [%d] for leave is an intermediate node => ignored!\n", 
			reqNodeID);
    return FALSE;
  }

  // If q is the mcSource, ignore the request.
  if (session->GetSrcNode() == reqNodeID)  
  {
    printf("Requesting node [%d] = mcSource node => ignored! \n", reqNodeID);
    return FALSE;
  }

  // Process the leave request
  BOOL retBool = ProcessLeaveRequest(session, userRequest);
  if (retBool == TRUE)
    printf("Successful leave. \n");
  return retBool;
}  // end StartLeaveAlgorithm


////////////////////////////////////////////////////////////////////////////////

BOOL Algorithm::ProcessLeaveRequest(Session* session, UserReq* userRequest)
{
#ifdef FUNCTION_TRACE
  fprintf(stderr, "++++++++ ProcessLeaveRequest() \n");
#endif

  int reqNodeID = userRequest->GetNodeID();  // requesting node ID  
  OnePath* backPath = new OnePath;
  int retBool;

  // Find the paths from mcSource to the other nodes in the current MC tree
  Path* allSharedPaths = session->FindPathFromSrcToDest();
	// a new Path object is alloc in this method => should dealloc it later. 

  // Find the path from mcSource to requesting node
  OnePath* reqPath = allSharedPaths->FindOnePathEndWith(reqNodeID);

  // Generate backward path from requesting node to mcSource node
  // in the MC tree (Note: start node of a path is not stored
  // in the list of nodes of the path)
  // Ex : source node [4] to requesting node [5]
  //	  reqPath  : -> ([4]) -> [3] -> [6] -> [5]
  //	  backPath : -> ([5]) -> [6] -> [3] -> [4]
  //
  backPath->GenBackPath(reqPath);

  // --------------------------------------------------------------------------------
  // If q is not a leaf node of the tree (i.e. has adjacent edge(s) from
  // below), update q's status (from LEAF_NODE to INTREE), and bandwidth
  // of edges on the path from q back to mcSource.  Do not delete any edges.
  //
  if (session->FindNode(reqNodeID)->GetEdgeNum(FORWARD) > 0) 
  {
    // 1. Update node status (from LEAF_NODE to INTREE)
    session->SetNodeStatus(reqNodeID, INTREE);   	     // intermediate node

     // 2. Update edge status    
    int startNodeID = backPath->GetPathNode()->GetNodeID();  // node [6] in above ex.
    Edge* edge = session->FindEdge(startNodeID, reqNodeID);  // edge [6,5] in above ex.
    edge->SetEdgeStatus(INTREE);

    // 3. Do not delete any edges from the MC tree
    // 4. Remove the path from list of treePaths
    retBool = RemovePathFromTreePaths(reqNodeID, session);

    // 5. Update bandwidth of edges on the path from q back to mcSource
    retBool = UpdatePathBandwidthOnLeave(session, backPath, reqNodeID);

    printf("Requesting node has adjacent edge(s) from below => no edges deleted.\n");
    delete allSharedPaths;
    delete backPath;
    return TRUE;
  }  // end if
  // --------------------------------------------------------------------------------

  // --------------------------------------------------------------------------------
  // q is a participant node and has no adjacent edge(s) from below.
  // Check and delete edges on the backPath (from q to mcSource).
  // The deletion stops when encountering a node that is a participant,
  // or has adjacent edge(s) from below.
  //
  int curNode  = reqNodeID;		// n = requesting node ID; node [5] in above ex.
  PathNode* preNodePtr = backPath->GetPathNode(); 
				// pn = previous node of n; node [6] in above ex.

  while (preNodePtr != NIL)	// while (n != mcSource) traverse the backPath
  { 
#ifdef FUNCTION_TRACE
  fprintf(stderr, "++++++++ current node [%d] in ProcessLeaveRequest() \n", curNode);
#endif

    int preNode = preNodePtr->GetNodeID();   	// previous node on backPath
    Node* preNodeObj = session->FindNode(preNode); 
	// get node object (in MC tree) from node ID

    // If [(n == requesting node) OR (n == intermediate node)]
    //    AND [n has no adjacent edge(s) from below]
    //
    if ((curNode == reqNodeID || session->GetNodeStatus(curNode) == INTREE) 
	&& (session->FindNode(curNode)->GetEdgeNum(FORWARD) == 0))
    {
      // 1. Update node status
      session->SetNodeStatus(curNode, UNSEEN);  

      // 2. No need to update edge status since the edge will be removed from MC tree.
      // 3. Remove edge (pn, n) from the MC tree 
      preNodeObj->RemoveAdjEdge(FORWARD, curNode);

      // 4. Remove the path from list of tree paths
      if (curNode == reqNodeID)
        retBool = RemovePathFromTreePaths(curNode, session);
    
      // 5. Add session bandwidth on edge (pn, n) to available bandwidth
      retBool = FreeSessionBandwidthOnEdge(preNode, curNode, session);
    } 
    // Else [(n != requesting node) AND (n == participant node)]
    //	    OR [n has adjacent edge(s) from below]
    //
    else if ((curNode != reqNodeID && session->GetNodeStatus(curNode) == LEAF_NODE)  
	     || (session->FindNode(curNode)->GetEdgeNum(FORWARD) > 0))
    {
      // Update bandwidth of edges on the path from n back to mcSource
      retBool = UpdatePathBandwidthOnLeave(session, backPath, curNode);

      // And stop the deletion
      break; 
    }
    else 
    {
      fprintf(stderr, "Fatal error: unreachable code in ProcessLeaveRequest()!\n");
      exit(1);
    }

    curNode = preNode;			  // n  = pn;
    preNodePtr = preNodePtr->GetNext();	  // pn = previous node of n;
  }  // end while

  delete allSharedPaths;
  delete backPath;
  return TRUE;
  // --------------------------------------------------------------------------------
}  // end ProcessLeaveRequest

////////////////////////////////////////////////////////////////////////////////

// 
// Update bandwidth of the edges on "backPath", from "stopNode" back to mcSource node
//
BOOL Algorithm::UpdatePathBandwidthOnLeave(Session* session, OnePath* backPath, 
						int stopNode)
{
#ifdef FUNCTION_TRACE
  fprintf(stderr, "++++++++++ UpdatePathBandwidthOnLeave() (path [%d -> %d])\n",
			stopNode, backPath->GetDestNode());
#endif

  int sessionID = session->GetSessID();
  PathNode* pnode = backPath->GetPathNode();
  int n = stopNode;
  if (stopNode != backPath->GetSrcNode())    // stopNode != requestingNode
  {
    while(pnode->GetNodeID() != stopNode)
      pnode = pnode->GetNext();

    // When we get here (pnode->GetNodeID == stopNode).
    // We want to get to previous node of stopNode n.  So ...
    pnode = pnode->GetNext();	
  }

  Path* allTreePaths = session->GetTreePaths();

  while (pnode != NIL)  		// equivalent to "while(n != mcSource)"
  {
    int nodeReqBw;  // bandwidth that was requested by node n
    int pn = pnode->GetNodeID();

    // Get bandwidth that was requested by node n.
    // (An intermediate node is considered to have requested bandwidth = 0)
    //
    if (session->GetNodeStatus(n) == INTREE)  // n is an intermediate node
      nodeReqBw = 0;
    else	// n is a participant
    {
      OnePath* nPath = allTreePaths->FindOnePathEndWith(n);
      if (nPath == NIL)		// impossible to get here
      {
	fprintf(stderr, "Fatal error: path node found in UpdatePathBandwidthOnLeave!\n");
	exit(1);
      }
      nodeReqBw = nPath->GetDestNodeReqBand();
    }  // end else	

    int newUsedBw = nodeReqBw;

    Node* treeNode = session->FindNode(n);
    EdgeItem* treeEdgeItem;
    // For every edge going out from n (in the MC tree)
    for (treeEdgeItem = treeNode->GetEdgeItem();
	 treeEdgeItem != NIL;
	 treeEdgeItem = treeEdgeItem->GetNext())
    {
      // Get end node ID of the edge
      int endNode = treeEdgeItem->GetEdge()->GetTargetNodeID();

      // Get the edge object from network graph (to access session used bandwidth)
      Edge* netwkEdge = data_mgr->GetGraph()->FindEdge(n, endNode);

      // Get session used bandwidth
      int bwAdjEdge = netwkEdge->GetSessionUsedBandwidth(sessionID);

      // Get the maximum required bandwidth
      newUsedBw = maxValue(newUsedBw, bwAdjEdge);
    }  // end for (treeEdgeItem)

    // Compare newUsedBw with current used bandwidth
    Edge* curNetwkEdge = data_mgr->GetGraph()->FindEdge(pn, n);
    int curUsedBw = curNetwkEdge->GetSessionUsedBandwidth(sessionID);
    if (newUsedBw > curUsedBw)
    {
      fprintf(stderr, "Fatal error with bandwidth in UpdatePathBandwidthOnLeave!\n");
      exit(1);
    }
    
    if (newUsedBw < curUsedBw)
    {
      // Update session used bandwidth on curNetwkEdge 
      // and available bandwidth of curNetwkEdge (in network graph)
      int retBool = curNetwkEdge->UpdateUsedBand(EDGE_DEC, sessionID, 
						(curUsedBw - newUsedBw));

      // Update available bandwidth of edge object in MC tree
      session->UpdateEdgeData(pn, curNetwkEdge);
    }
    else  	// (newUsedBw == curUsedBw) 
      break;  	// stop the update as soon as there is no change 
		// (it would be useless to continue)

    // Advance to next node on backPath
    n = pn;
    pnode = pnode->GetNext();
  }  // end while (pnode) 
 
  return TRUE;
}  // end UpdatePathBandwidthOnLeave

////////////////////////////////////////////////////////////////////////////////

//
// Delete session used bandwidth on edge(startNode, endNode)
//
BOOL Algorithm::FreeSessionBandwidthOnEdge(int startNode, int endNode, Session* session)
{
#ifdef FUNCTION_TRACE
  fprintf(stderr, "++++++++++ FreeSessionBandwidthOnEdge() (edge [%d -> %d])\n",
			startNode, endNode);
#endif

  int sessionID = session->GetSessID();
  int retValue;

  // Update edge object in network graph
  // Get the edge object of edge(startNode, endNode)
  Edge* edge = data_mgr->GetGraph()->FindEdge(startNode, endNode);
  if (edge == NIL)
  {
    fprintf(stderr, "Fatal error: NIL edge in method DeleteEdgeBandwidth()!\n");
    return FALSE;
  }

  // Get the session used bandwidth on this edge
  int sessionUsedBandwidth = edge->GetSessionUsedBandwidth(sessionID);

  // Update session used bandwidth and edge available bandwidth (current_band)
  retValue = edge->UpdateUsedBand(EDGE_DEC, sessionID, sessionUsedBandwidth);
 
#ifdef DEBUG_LEVEL_1
  if (retValue != 0)
  {
    fprintf(stderr, "Wrong updated bandwidth in method FreeSessionBandwidthOnEdge!\n");
    exit(1);
  }
#endif
 
  // Update edge object in the MC tree (update "current_band" and "total_band")
  session->UpdateEdgeData(startNode, edge);

  return TRUE;
}  // end FreeSessionBandwidthOnEdge


////////////////////////////////////////////////////////////////////////////////

BOOL Algorithm::RemovePathFromTreePaths(int leaveNode, Session* session)
{
#ifdef FUNCTION_TRACE
  fprintf(stderr, "++++++++++ RemovePathFromTreePaths() (leave node [%d])\n",
			leaveNode);
#endif

  // Update numDependants[] of the nodes on the path to be removed
  OnePath* removedPath = session->GetTreePaths()->FindOnePathEndWith(leaveNode);
  if (removedPath != NIL)
    session->UpdateNumDependants(removedPath, -1);  // -1 for leave (1 for join)
  else
  {
    fprintf(stderr, "Fatal Error in RemovePathFromTreePaths(): path to be removed = NIL!\n");
    exit(1);  
  }

  // Actually remove the path from treePaths
  BOOL retBool = session->GetTreePaths()->RemovePath(leaveNode);
  if (retBool == FALSE)
    fprintf(stderr, "Fatal Error in RemovePathFromTreePaths(): should not get here!\n");

  return retBool;  
}


////////////////////////////////////////////////////////////////////////////////

void Algorithm::DisplayCandidateS2DPaths(Path* candidatePaths, Session* session,
						UserReq* userRequest)
{
#ifdef FUNCTION_TRACE
  fprintf(stderr, "++++++++++ DisplayCandidateS2DPaths() (calls JoinCost()) \n");
#endif

  OnePath* newPath;
  OnePath* sharedPath;
  OnePath* joinPath;
  int      connectionNode;
  COST 	   retCosts(0, 0, 0);

  if (candidatePaths == NIL) return;

  // Find the paths from mcSource to the other nodes in the current MC tree
  Path* allSharedPaths = session->FindPathFromSrcToDest();

  // Among all candidate paths, select the best S2D path.
  for(newPath  = candidatePaths->GetPath(); 
      newPath != NIL;
      newPath  = newPath->GetNext())
  {
    // Get the connection node ID
    connectionNode = newPath->GetSrcNode();

    // Find the shared path from mcSource to connection node
    sharedPath = allSharedPaths->FindOnePathEndWith(connectionNode);

    printf("C[%.4f] ", JoinCost(sharedPath, newPath, session, userRequest));
    joinPath = Combine2Paths(sharedPath, newPath);
    joinPath->Display();
    delete joinPath;
  }  // end for(newPath)

  delete allSharedPaths;
  return;
}  // end DisplayCandidateS2DPaths



////////////////////////////////////////////////////////////////////////////////
// 
// Notes:
// - 5 steps to add/remove a node to/from the MC tree:
//	1. Update node status
//	2. Update edge status in MC tree (*)
//	3. Add/RemoveAdjEdge
//	4. Add/RemovePathTo/FromTreePaths
//	5. Update bandwidth
//
// 	(*) Edge status in the MC tree is used only for "display".  No other
//	methods use this information.  So we do not need to update edge
//	status.  However, to display correct information of the tree
//	for debugging purposes, we should update edge status.
//
////////////////////////////////////////////////////////////////////////////////