#include "alg.h"

//////////////////////////////////////////////////////////////////////////////
//
//	Method name : ShortestPath
//
// 	Description :
// 		Method to find the shortest path from start_node to every 
//		other node in the network graph (Dijkstra's algorithm).
//
//	Parameter :
// 		 - ADJ_TYPE type : Adjacency list type to find the path
//			FORWARD  => Used to maintain graph information
//	               	BACKWARD => Used to find the shortest path from 
//			   requesting node to every other node in network graph
// 		- int start_node
// 		- COST_TYPE cost_type : condition type to find the paths 
//			(WEIGHT, DELAY, CONGESTION)
//
//	Return : Path*
//		 List of paths from start_node to every other node in network graph 
//
//	Note : When running Dijkstra's, "start_node" is the requesting node,
//             and "type" is BACKWARD.
//
//      Note : The status of a node i in this method:
//      - UNSEEN : i has never been visited (weight[i] = delay[i] = 0)
//      - INTREE : i was visited, and the shortest path from start_node to i has
//                 been established (weight[i] = delay[i] = cost of the shortest path)
//      - FRINGE : i has been visited, but the shortest path from start_node to
//                 i has not been established yet 
//                 (weight[i] = delay[i] = current cost >= best cost)
//
//////////////////////////////////////////////////////////////////////////////
Path*	Algorithm::ShortestPath(ADJ_TYPE type, int start_node, COST_TYPE cost_type,
					UserReq* userRequest)
{
#ifdef FUNCTION_TRACE
  fprintf(stderr, "++++++++++ ShortestPath() \n");
#endif

	int 	reqBand   = userRequest->GetQos().GetBandwidth();
	int	tot_nodes = data_mgr->GetGraph()->GetNodeNum();
	                    // total number of nodes in network graph
	int	*status = new int[tot_nodes];
	int	*parent = new int[tot_nodes];
	int	*weight = new int[tot_nodes];
	int	*delay  = new int[tot_nodes];
	float	*congestion = new float[tot_nodes];

	EdgeItem	*edge_list, *e_tmp;

	int	src_node,dest_node;
	int	found_node_num;     // records number of INTREE nodes
	int	fringeNum = 0;      // records number of FRINGE nodes
	        // number of UNSEEN nodes = tot_nodes - (found_node_num + fringeNum)
	BOOL	struck = FALSE;
	int 	minNode;

	for (int i = 0; i<tot_nodes; i++) {
		status[i] = UNSEEN;
		parent[i] = -1;
		weight[i] = -1;
		delay[i] = -1;
		congestion[i] = -1.0;
	}
	found_node_num = 0;
	src_node = start_node;  // = requesting node in Dijkstra's algorithm
	status[src_node] = INTREE;
	weight[src_node] = 0;
	delay[src_node] = 0;
	congestion[src_node] = 0.0;
	parent[src_node] = -1;  // start node does not have a parent
	found_node_num++;

	// Keep finding path until the number of found nodes is same as 
	// the total number of nodes in network graph, or stuck in a cycle.
	//
 	while(found_node_num < tot_nodes && struck==FALSE) {

		edge_list = data_mgr->GetGraph()->FindEdgeHead(type, src_node);
		          // get pointer to the fw/bw adj list of "src_node"
		e_tmp = edge_list;	

		// Update FRINGE nodes & cost
		while(e_tmp != NIL) {  // for every node/edge adjacent to "src_node"
			dest_node = e_tmp->GetEdge()->GetTargetNodeID();
			       // = a node adjacent to "src_node"
			if (status[dest_node] == FRINGE && cost_type == WEIGHT) {
			   if (weight[src_node] + e_tmp->GetEdge()->GetWeight() < weight[dest_node]) {
			    parent[dest_node] = src_node;
			    weight[dest_node] = weight[src_node]+e_tmp->GetEdge()->GetWeight();
			    delay[dest_node] = delay[src_node]+e_tmp->GetEdge()->GetDelay();
			    congestion[dest_node] = congestion[src_node] + 
				e_tmp->GetEdge()->CongestionInc(e_tmp->GetEdge()->GetEdgeUsedBand(), reqBand);
			  };
			} else if (status[dest_node] == FRINGE && cost_type == DELAY) {
			   if (delay[src_node] + e_tmp->GetEdge()->GetDelay() < delay[dest_node]) {
			    parent[dest_node] = src_node;
			    weight[dest_node] = weight[src_node]+e_tmp->GetEdge()->GetWeight();
			    delay[dest_node] = delay[src_node]+e_tmp->GetEdge()->GetDelay();
			    congestion[dest_node] = congestion[src_node] + 
				e_tmp->GetEdge()->CongestionInc(e_tmp->GetEdge()->GetEdgeUsedBand(), reqBand);
			  };
			} else if (status[dest_node] == FRINGE && cost_type == CONGESTION) {
			   if (congestion[src_node] + 
				e_tmp->GetEdge()->CongestionInc(e_tmp->GetEdge()->GetEdgeUsedBand(), reqBand)
				< congestion[dest_node]) {
			    parent[dest_node] = src_node;
			    weight[dest_node] = weight[src_node]+e_tmp->GetEdge()->GetWeight();
			    delay[dest_node] = delay[src_node]+e_tmp->GetEdge()->GetDelay();
			    congestion[dest_node] = congestion[src_node] + 
				e_tmp->GetEdge()->CongestionInc(e_tmp->GetEdge()->GetEdgeUsedBand(), reqBand);  
			  };
			}

			if (status[dest_node] == UNSEEN) {   // unvisited
			        status[dest_node] = FRINGE;  // just visisted
				fringeNum++;
				parent[dest_node] = src_node;
				weight[dest_node] = weight[src_node] + e_tmp->GetEdge()->GetWeight();
				delay[dest_node] = delay[src_node] + e_tmp->GetEdge()->GetDelay();
				congestion[dest_node] = congestion[src_node] + 
				   e_tmp->GetEdge()->CongestionInc(e_tmp->GetEdge()->GetEdgeUsedBand(), reqBand);
			}
			e_tmp = e_tmp->GetNext();  // e_tmp = e_tmp->next;
		}  // end while(e_tmp)

		// Find node with minimum cost among the FRINGE nodes
		// (to make one of the FRINGE nodes INTREE)
		//
		if (fringeNum == 0) 
		        struck = TRUE;         // get stuck in a cycle
		else {
		        minNode = start_node;  // start_node acts as a minimal value;
			                       // we could use "minNode = -1;"
			for (int i=0; i<tot_nodes; i++) {
			  // consider only FRINGE (visited) nodes
			        if (status[i] == FRINGE) {  // visited node
					if (minNode == start_node) minNode=i;
					else if (cost_type == WEIGHT && weight[minNode] > weight[i]) 
						minNode = i;
					else if (cost_type == DELAY && delay[minNode] > delay[i]) 
						minNode = i;
					else if (cost_type == CONGESTION && congestion[minNode] > congestion[i]) 
						minNode = i;
				}
			}
			src_node = minNode;       // minNode becomes new src_node
			status[src_node]=INTREE;  // shortest path established
			fringeNum--;              // one less FRINGE node
			found_node_num++;         // one more INTREE node
		}
	} // end while(found_node_num)

	// When we get here, the Dijkstra's algorithm is completed.
	// For every node i, weight[i] (delay[i]) contains the weight (delay) 
	// of the shortest path from start_node to i.

	// Generate list of paths from parent[], status[], weight[]
	// delay[] information
	Path* path = new Path(start_node, type);

	for (int leaf_node=0; leaf_node<tot_nodes; leaf_node++) {
	  // for every node in network graph

		if (leaf_node == start_node) continue;
		if (parent[leaf_node] == -1) continue;

		// Construct the path from start_node to leaf_node (a list of 
		// nodes).  Note: start_node is not recorded in this list.
		// Go backward from leaf_node to start_node using parent[] array.
		// weight[leaf_node], delay[leaf_node] contain the weight, delay
		// of the shortest path from start_node to leaf_node.
		// (That's why AddNodeToHead need not accumulate the 
		// weights/delays of edges on the shortest path)
		//
		OnePath* one_path = new OnePath(start_node, weight[leaf_node], delay[leaf_node]);
		   // paths do not store congestion cost
		one_path->AddNodeToHead(leaf_node);
		int parent_node = parent[leaf_node];
		while (parent_node != start_node) {
			one_path->AddNodeToHead(parent_node);
			parent_node = parent[parent_node];
		}

		// Add the path just constructed to list of paths
		path->AddOnePath(one_path);
	}

	delete [] status;
	delete [] parent;
	delete [] weight;
	delete [] delay;
	return path;     // list of shortest paths
}


//////////////////////////////////////////////////////////////////////////////
//
//	Method name : UpdateTreeWithPath
//
// 	Description :
//		Method to add and update the multicasting tree using 
//		found paths (shared path and new path)
//
//	Parameter :
//		- OnePath* old_one_path : a path from the mcSource of a 
//		  	session to a connection node in MC tree (shared path)
//		- OnePath* new_onepath : a path from the connection node in the 
//		 	multicasting tree to the requesting node (new path)
//		- Session* session : session to be used by algorithm
//
//	Return : BOOL (always return TRUE)
//
//////////////////////////////////////////////////////////////////////////////
BOOL	Algorithm::UpdateTreeWithPath(OnePath* shared_onepath, OnePath* new_onepath, Session* sess)
{
#ifdef FUNCTION_TRACE
  fprintf(stderr, "++++++++++ UpdateTreeWithPath() \n");
#endif

	Node* 	nd;
	Edge* 	n_edge;
	int 	src_node=-1, dest_node=-1;
        PathNode* pnode;
	COST curCost(0,0,0);

	if (new_onepath == NIL) {
		printf( "FILE[%s], LINE[%d] : NULL path\n",
			__FILE__, __LINE__);
		exit(1);
	}	

        pnode = new_onepath->GetPathNode();   // new_onepath->node_list
	src_node = new_onepath->GetSrcNode(); // src_node here is the connection
	                                      // node (not stored in node_list)
	while(pnode != NIL) {  // traverse from connection node to requesting node
		dest_node = pnode->GetNodeID();	

		if (sess->GetNodeStatus(dest_node) == UNSEEN) {  // not in tree
			STATUS status;

			if (dest_node == new_onepath->GetDestNode())
			        status = LEAF_NODE;  // participant node
			else 	status = INTREE;     // intermediate node

			n_edge = new Edge;  // alloc a new Edge object

			// n_edge->SetSrcNode(nd);  // Jin's code -> wrong
			n_edge->SetTargetNodeID(dest_node);
			n_edge->SetEdgeStatus(status);  
			sess->SetNodeStatus(dest_node, status);
			      // sess->in_session[dest_node] = status;
			Node* nd = sess->FindNode(src_node);
			n_edge->SetSrcNode(nd);
			nd->AddAdjEdge(FORWARD, n_edge);  
			      // add n_edge to adj list of "nd"

			// Accumulate costs of current path (from mcSource to dest_node)
			Edge* d_edge = data_mgr->GetGraph()->FindEdge(src_node, dest_node);
			curCost.weight += d_edge->GetWeight();		
			curCost.delay  += d_edge->GetDelay();

			// Store costs in n_edge
			n_edge->SetWeight(curCost.weight);
			n_edge->SetDelay(curCost.delay); 

			// Note: hopCount of a path will be accumulated in methods
			// AddNodeToHead() or FindCost().

		}  // end if
		// else if loops are not allowed, we can never get here;
		// else (loops are allowed), we may have to do something here.
			
		src_node = dest_node;	
		pnode = pnode->GetNext();  // pnode = pnode->next;
	}  // end while

	return TRUE;
}

// Note: 
// - n_edge must store the costs of the whole path because this info cannot
//   be kept anywhere else.
// - If we represent the MC tree as a list of paths from mcSource to the other
//   nodes of the tree, then the costs of each path can be stored with the path
//   itself (attributes weight_sum and delay_sum in class OnePath).


//////////////////////////////////////////////////////////////////////////////
//
//	Method name : UpdateBandwidth
//
// 	Description : For every edge e on path "one_path",
// 	   - update "current_band" of e in network graph, and used bandwidth
//           of current session (in network graph object)
// 	   - update "current_band" of e in the multicasting tree 
//
//	Parameter :
//		- PATH_TYPE pathType: shared path or new path
//		- OnePath* one_path : a path containing edges to be updated
//		- Session* session  : session to be used by algorithm
//		- UserReq* user_req : user request
//
//	Return : BOOL 
//
//	Note : 
//		- Only network GRAPH has full information on bandwidths of edges
//                (i.e. available bandwidth, and used bandwidth of each session)
//		- Multicasting tree of a session stores only available
//                bandwidths of edges.
//
//////////////////////////////////////////////////////////////////////////////
BOOL	Algorithm::UpdateBandwidth(PATH_TYPE pathType, OnePath* one_path, 
					Session* sess, UserReq* user_req)
{
#ifdef FUNCTION_TRACE
  fprintf(stderr, "++++++++++ UpdateBandwidth() \n");
#endif

	int	req_band = user_req->GetQos().GetBandwidth();
	int	req_node = user_req->GetNodeID();
	int	req_sess_id = user_req->GetSessionID();
	int	src_node, dest_node;
	int	ret;

	// Invalid requested bandwidth
	if (req_band <= 0) return TRUE;

	if (one_path == NIL) return TRUE;

	PathNode* pn = one_path->GetPathNode();  
			// pn = one_path->node_list; (list of nodes on the path)
	src_node = one_path->GetSrcNode();  
			// get start node of the path (not stored in node_list)

	while (pn != NIL) {
	        dest_node = pn->GetNodeID();  // end node ID
		//
		// Note: the same edge connecting start node and end node has
		// 2 objects associated with it, one in the network graph and
		// the other in the multicast tree (sess->mc_tree). So we 
		// update bandwidth info in both objects.
		//
		// Update edge object in the network graph
		Edge* d_edge = data_mgr->GetGraph()->FindEdge(src_node, dest_node);
		if (d_edge == NIL) {
			fprintf(stderr, "FILE[%s], LINE[%d] : fatal error !!\n", 
				__FILE__, __LINE__);
			exit(1);
		}
		switch (pathType)
		{
		  // Update attribute "current_band" and used bandwidth of 
		  // req_sess_id (in the list "used_band_head")
		  //
		  case NEW_PATH:
		    ret = d_edge->UpdateUsedBand(EDGE_ADD, req_sess_id, req_band);
		    break;
		  case SHARED_PATH:
		    ret = d_edge->UpdateUsedBand_SharedPath(req_sess_id, req_band);
		    break;
		  default:
		    fprintf(stderr, "Invalid path type (method UpdateBandwidth)!\n");
		    return FALSE;
		}  //  end switch

		if (ret < 0) {
			fprintf(stderr, "Invalid UpdateBandwidth request !\n");
			return FALSE;
		}

		// Update edge object in the MC tree.
		// if (sess->GetNodeStatus(dest_node) == LEAF_NODE) 
		sess->UpdateEdgeData(src_node, d_edge);
		   // update attributes "total_band" and "current_band"

		src_node = dest_node;
		pn = pn->GetNext();  // pn = pn->next;
	}

	return TRUE;
}

//////////////////////////////////////////////////////////////////////////////
//
//	Method name : IsValidQOS
//
// 	Description :
//		Method to validate requested QoS of edges on a path
//
//	Parameter :
//		- OnePath* old_one_path : a path from a source node of a 
//		  	session to connection node (A) in the multicasting tree
//		- OnePath* new_onepath : a path from connection node (A) in the 
//		 	multicasting tree to requesting node
//		- Session* session : session to be used by algorithm
//
//	Return : TRUE if requested QoS is satisfied; FALSE otherwise.
//
//////////////////////////////////////////////////////////////////////////////
BOOL 	Algorithm::IsValidQOS(OnePath* shared_onepath, OnePath* new_onepath, 
				Session* sess, UserReq* user_req)
{
#ifdef FUNCTION_TRACE
  fprintf(stderr, "++++++++++++++ IsValidQOS() \n");
#endif

	int	cur_hop;
	int	cur_delay;
	COST	sharedCosts(0, 0, 0);
	COST	newCosts(0, 0, 0);

	if (shared_onepath != NIL)
	{
	  sharedCosts = shared_onepath->FindCost(data_mgr->GetGraph(), NEW_PATH);
	  shared_onepath->SetCosts(sharedCosts);
	}
	
	if (new_onepath != NIL)
	{
	  newCosts = new_onepath->FindCost(data_mgr->GetGraph(), NEW_PATH);
	  new_onepath->SetCosts(newCosts);
	}

	// Get delay sum and path length of the combined path
	if (shared_onepath != NIL && new_onepath != NIL) { 
		cur_hop = shared_onepath->GetHop() + new_onepath->GetHop();
		cur_delay = shared_onepath->GetDelay() + new_onepath->GetDelay();
	} else if (new_onepath != NIL) {     // and no shared path 
		cur_hop = new_onepath->GetHop();
		cur_delay = new_onepath->GetDelay();
	} else if (shared_onepath != NIL) {  // and no new path
		cur_hop = shared_onepath->GetHop();
		cur_delay = shared_onepath->GetDelay();
	} else {  // shared path and new path are both NIL
		fprintf(stderr, "Fatal error in IsValidQOS(): should not get here!\n");
		return FALSE;	
	}

	// Check requested path length
	if (cur_hop > user_req->GetQos().GetHop()) 
	{
#ifdef DEBUG_LEVEL_1
		printf("Exceed requested path length (connection node");
		if (new_onepath != NIL)
		   printf("[%d]) on path: \n", new_onepath->GetSrcNode());
		else printf("[%d]) on path: \n", shared_onepath->GetDestNode());
		OnePath* joinPath = Combine2Paths(shared_onepath, new_onepath);
		joinPath->Display();
		delete joinPath;
#endif
		return FALSE;
	}

	// Check requested delay
	if (cur_delay > user_req->GetQos().GetDelay()) 
	{
#ifdef DEBUG_LEVEL_1
		printf("Exceed requested delay (connection node");
		if (new_onepath != NIL)
		   printf("[%d]) on path: \n", new_onepath->GetSrcNode());
		else printf("[%d]) on path: \n", shared_onepath->GetDestNode());
		OnePath* joinPath = Combine2Paths(shared_onepath, new_onepath);
		joinPath->Display();
		delete joinPath;
#endif		
		return FALSE;
	}

	// Check requested bandwidth 
	// (If the path is NIL, CheckBandwidth() returns TRUE)
	//
	BOOL ret;
	ret = CheckBandwidth(SHARED_PATH, shared_onepath, sess, user_req);
	if (ret == FALSE) {
#ifdef DEBUG_LEVEL_1
	   	printf("Insufficient bandwidth on shared path (connection node [%d]):\n", 
				shared_onepath->GetDestNode()); 
	   	shared_onepath->Display();		
#endif
		return FALSE;
	}

	ret = CheckBandwidth(NEW_PATH, new_onepath, sess, user_req);
	if (ret == FALSE) {
#ifdef DEBUG_LEVEL_1
		printf("Insufficient bandwidth on new path (connection node [%d]):\n", 
				new_onepath->GetSrcNode()); 
		new_onepath->Display();
#endif
		return FALSE;
	}

	return TRUE;
}

// Note:
// - shared_onepath: 
//    + delay_sum (GetDelay()) is got from "connection edge"
//		(see mc_tree and Session::FindPathFromSrcToDest())
//    + hop count (GetHop()) is accumulated by AddNodeToHead()
//		(see Session::FindPathFromSrcToDest() and OnePath::AddNodeToHead())
// - new_onepath:
//    + delay_sum (GetDelay()) is got from delay[] array of Dijkstra's algorithm
//		(see ShortestPath())
//    + hop count (GetHop()) is accumulated by AddNodeToHead()
//		(see ShortestPath() and OnePath::AddNodeToHead())


//////////////////////////////////////////////////////////////////////////////
//
//	Method name : CheckBandwidth
//
// 	Description :
//		Check the bandwidth of every edge in the path to see
//		whether requested bandwidth exceeds current bandwidth of the edge
//
//	Parameter :
//		- PATH_TYPE pathType: shared path or new path
//		- OnePath* one_path : a path containing edges to be checked
//		- Session* session : session to join (not used for now)
//		- UserReq* user_req : user request
//
//	Return : BOOL
//		- TRUE : enough bandwidth (or the path is NIL)
//		- FALSE : exceed the bandwidth on one of the edges on the path
//
//////////////////////////////////////////////////////////////////////////////
BOOL	Algorithm::CheckBandwidth(PATH_TYPE pathType, OnePath* one_path, 
					Session* sess, UserReq* user_req)
{
#ifdef FUNCTION_TRACE
  fprintf(stderr, "++++++++++++++++ CheckBandwidth() \n");
#endif

	int	req_band = user_req->GetQos().GetBandwidth();
	// int	req_node = user_req->GetNodeID();
	int	req_sess_id = user_req->GetSessionID();
	int	src_node, dest_node;


	if (req_band <= 0) return TRUE;
	if (one_path == NIL) return TRUE;

	PathNode* pn = one_path->GetPathNode();  // pn = one_path->node_list;
	src_node = one_path->GetSrcNode();  // src_node = start node of the path;

	while(pn != NIL) {
		dest_node = pn->GetNodeID();	
		
		Edge* d_edge = data_mgr->GetGraph()->FindEdge(src_node, dest_node);
		               // find the edge connecting src_node and dest_node
		if (d_edge == NIL) {
			fprintf(stderr, "FILE[%s], LINE[%d] : invalid pointer !!\n", 
				__FILE__, __LINE__);
			fprintf(stderr, "Edge : from [%d] -> to [%d]\n", 
				src_node, dest_node);
			one_path->Display(stderr);
			exit(1);
		}
		
		switch(pathType)
		{
		  case NEW_PATH:
		    if (d_edge->IsEnoughBand(req_band) == FALSE) 
			return FALSE;
		    break;
		  case SHARED_PATH:
		    if (d_edge->IsEnoughBand_SharedPath(req_band, req_sess_id) == FALSE)
			return FALSE;
		    break;
		  default:
		    fprintf(stderr, "Invalid path type (method CheckBandwidth)!\n");
		    return FALSE;
		}  // end switch

		src_node = dest_node;
		pn = pn->GetNext();    // pn = pn->next;
	}
	return TRUE;
}
