///////////////////////////////////////////////////////////////////////////////
//
//	File name : edge.cc			Last updated: Feb 24, 98
//
//	Description :
//		This files have "edge" related class : EdgeIem, Edge
//
//
//	Authors: Jin-Hwee Park / U.T.
// 	Date   : Jan-22-1998
//
///////////////////////////////////////////////////////////////////////////////

#include "edge.h"
#include "graph.h"

//
// Copy the attributes of parm "used" to current object.
//
UsedBand::UsedBand(const UsedBand* used) {
	session_id = used->GetSessID();
	used_bandwidth = used->GetBand();
	next = NIL;
}

Edge::Edge() {
	source_node = NIL;
	target_node_id = -1;
	total_band = 0;
	current_band = 0;
	edge_weight = 0;
	edge_delay = 0;
	used_band_head = NIL;
	edge_status = INTREE;  // for a new edge in an MC tree
	edgeNormFactor = 1.0;
}

Edge::Edge(Node* src, int t_node, int weight, int delay, int t_band) {
	source_node = src;
	target_node_id = t_node;
	total_band = t_band;
	current_band = t_band;
	edge_weight = weight;
	edge_delay = delay;
	used_band_head = NIL;
	edge_status = UNSEEN;  // for a new edge in network graph
	edgeNormFactor = 1.0;
}

//
// Copy the attributes of parm "edge" to current object
//
Edge::Edge(const Edge* edge)
{
	if (edge  == NIL) {
		printf("FILE[%s], LINE[%d] : Invalid edge pointer \n",
			__FILE__, __LINE__);
		exit(1);
	}
	SetSrcNode(edge->GetSrcNode());
	SetTargetNodeID(edge->GetTargetNodeID());
	SetWeight(edge->GetWeight());
	SetDelay(edge->GetDelay());
	SetTotBand(edge->GetTotBand());
	SetCurBand(edge->GetCurBand());
	edge_status = INTREE;		// meaning ?? why INTREE ??

	// Get the head of the list of used bandwidth items of "edge"
	UsedBand* used_band = edge->GetUsedBandHead();
	SetUsedBandHead(NIL);      // current_object.head = NIL
	while(used_band != NIL) {  // for every item in the list of "edge"
		// add a new item with same attributes to list of current_object
		AddUsedBand(used_band->GetSessID(), used_band->GetBand());
		used_band = used_band->GetNext();  // get next item
	}
}

Edge::~Edge() {
	UsedBand* u_band;

	// Free the whole list of used bandwidth items
	while(used_band_head != NIL) {
		u_band = used_band_head;
		used_band_head = used_band_head->GetNext();
			// used_band_head = used_band_head->next;
		delete u_band;
	}
}

//
// Allocate a new item (for session sess_id that requires bandwidth band).
// Add the new item to the front of the list of used bandwidth items
//
void    Edge::AddUsedBand(int sess_id, int band) {
	UsedBand* head = used_band_head;  	    // get head of the list
	UsedBand* new_band = new UsedBand(sess_id, band); // alloc new item

	new_band->SetNext(head);  // new_band->next = head;
	head = new_band;
}

//
// Update the used bandwidth on this edge of session sess_id.
// The requested or freed bandwidth is band.
// Returns the updated used bandwidth of sess_id.
//
int	Edge::UpdateUsedBand(ACTION_TYPE action, const int sess_id, const int band)
{
	UsedBand* u_band = used_band_head;  // head of list of used bandwidth items
	// UsedBand* pre_u_nband = NIL;     // unused
	int	ret = -1;

	// Update available bandwidth
	switch(action) {
		case EDGE_ADD:  // join request
			if (current_band < band)  // avail B < requested B
				return -1;
			DecCurBand(band);         // avail B -= requested B
			break;

		case EDGE_DEC:  // leave request
			if (u_band == NIL)	 // fatal error
				return -1;
			IncCurBand(band);	 // avail B += released B
			break;
		default:
			return -1;
			break;
	}

	// Update used bandwidth (item in the list of used bandwidth items)
	while(u_band != NIL) {		// for every item of the list
		if (u_band->GetSessID() == sess_id) {
			switch(action) {
				case EDGE_ADD:  // join request
				{
				ret =  u_band->IncBand(band);
				   // ret = new updated used bandwidth of sess_id
				break;
				}
				case EDGE_DEC:  // leave request
				{
				if (u_band->GetBand() < band) {
				  // current used bandwidth < bandwidth to be freed
					fprintf(stderr, "FILE[%s], LINE[%d] : Used bandwidth of edge problem !\n",
						__FILE__, __LINE__);
					exit(1);
				}
				ret = u_band->DecBand(band);
				}
			}
			return ret; // ret = updated used bandwidth
		}
		u_band = u_band->GetNext();   // get next item from list
	}

	// Leave request ( => session must exist) but cannot find
	// used bandwidth item of this session => fatal error.
	//
	if (action == EDGE_DEC) {
		fprintf(stderr, "FILE[%s], LINE[%d] : fatal error !\n",
			__FILE__, __LINE__);
		exit(1);
	}

	// If the action is EDGE_ADD and there is currently no item for requested session
	// Allocate a new item; set the used bandwidth to band.
	// Add new item to the front of the list of used bandwidth items.
	//
	u_band = new UsedBand(sess_id, band);  // alloc new item
	u_band->SetNext(used_band_head);       // u_band->next = head;
	used_band_head = u_band;
	return band;	    	// return updated used bandwidth
	// return current_band; // this line is Jin's code -> wrong
}

void	Edge::UpdateEdge(const Edge* edge) 
{
//	SetWeight(edge->GetWeight());
//	SetDelay(edge->GetDelay());
	SetTotBand(edge->GetTotBand());  // total_band = total_band
	SetCurBand(edge->GetCurBand());  // current_band = current_band
}

//
// Display the attributes and the list of used bandwidth items
//
void	Edge::Display(void) {
	// Print the attributes
	printf("* Edge [%d->%d] : W[%d], D[%d], TotBand[%d], CurBand[%d]\n",
		GetSrcNode()->GetNodeID(),
		 GetTargetNodeID(), GetWeight(), GetDelay(), GetTotBand(), GetCurBand());

	// Print the list of used bandwidth items
	UsedBand* u_band = GetUsedBandHead();
	while(u_band != NIL) {
		printf("* UsedBand : Session [%d], Band[%d]\t",
			u_band->GetSessID(), u_band->GetBand());
		u_band = u_band->GetNext();
	}
	printf("\n");
}

//
// Delete the whole list of used bandwidth items
//
void	Edge::DeleteUsedBand()
{
	SetCurBand(GetTotBand());  // update available bandwidth to total bandwidth capacity
	UsedBand* band = GetUsedBandHead();  // get head of list of used bandwidth items
	while(band != NIL) {		// for every item of the list
		UsedBand* tmp = band; 	// tmp = band;
		band = band->GetNext(); // band = band->next;
		delete tmp;
	}
	SetUsedBandHead(NIL); // set head pointer to NIL (empty list)
}

///////////////////////////////////////////////////////////////////////////

//
// Check bandwidth of an edge on a shared path for a join request.
// Returns TRUE if there's enough bandwidth; returns FALSE otherwise.
//
BOOL Edge::IsEnoughBand_SharedPath(const int req_band, const int sessionID)
{
  if (current_band == UN_LIMIT) return TRUE;
  if (req_band - GetSessionUsedBandwidth(sessionID) <= current_band)
    return TRUE;
  return FALSE;
}

//
// Given session ID, traverse the list of used bandwidth items,
// and return the bandwidth used by the given session.
// Precondition: the used bandwidth item of the given session
//		 must exist in the list.
//
int Edge::GetSessionUsedBandwidth(const int sessionID)
{
  UsedBand* itemPtr;
  for (itemPtr = used_band_head; itemPtr != NIL; 
	itemPtr = itemPtr->GetNext())
    if (itemPtr->GetSessID() == sessionID)
      return (itemPtr->GetBand());

  printf("Fatal error in GetSessionUsedBandwidth (session %d)!\n", sessionID);
  return (-1);
}

//
// Update the session used bandwidth and available bandwidth of the edge.
// Return the updated session used bandwidth.
// Precondition: the used bandwidth item of the given session
//		 must exist in the list of used bandwidth items.
//
int Edge::UpdateUsedBand_SharedPath(const int sessionID, const int req_band)
{
  UsedBand* itemPtr;

  for (itemPtr = used_band_head; 
       itemPtr != NIL; 
       itemPtr = itemPtr->GetNext())
    if (itemPtr->GetSessID() == sessionID)
    {
      int currentUsedBand = itemPtr->GetBand();
      if (req_band > currentUsedBand)
      { 
	// Update (decrease) available bandwidth of the edge
	DecCurBand(req_band - currentUsedBand);

	// Update session used bandwidth
	itemPtr->SetBand(req_band);  // itemPtr->used_bandwidth = req_band;
      }
      // else session has already got enough bandwidth on this edge => do nothing
      return (itemPtr->GetBand());  // return updated session used bandwidth
    }

  printf("Fatal error in UpdateUsedBand_SharedPath!\n");
  return (-1);
}


float Edge::Congestion(int usedBand)
{
/// Michalis made this C/(C-U) instead of 1/(C-U)
  if (usedBand < total_band) {
    return( (float) usedBand * total_band  /(float)(total_band - usedBand));
  };
  return INFINITY;
}


float Edge::CongestionInc(int usedBand, int reqBand)
{
/// Michalis changed this to return the weight of the edge
  // return (float) GetWeight();
   return(Congestion(usedBand + reqBand) - Congestion(usedBand));
}


///////////////////////////////////////////////////////////////////////////
//
// Updates:
// Feb 09: add methods IsEnoughBand_SharedPath, GetSessionUsedBandwidth, 
//		and UpdateUsedBand_SharedPath
// Feb 24: added attribute edgeNormFactor to class Edge (UT)
//
/////////////////////////////////////////////////////////////////////////////// 
