// path.h					Last updated: Feb 24, 98

#ifndef Path_H
#define Path_H

#include "common.h"
#include "graph.h"
#include "grp_req.h"


//
//	Class Name : PathNode
//
//	Description :
//		Representation of a node in the path
//
class PathNode {
private:
	int	node_id; // node ID
	PathNode* next;  // to form a path which is a list of PathNodes

public:
	PathNode(int n) {     // init the attributes
		node_id = n;
		next = NIL;
	}
	PathNode(PathNode* pn); // copy attributes of parm "pn" to current object
				// implementation in path.cc
	int	GetNodeID() const { return node_id; }; // return node ID
	void	SetNodeID(int id) { node_id = id; }    // set node ID to parm id

	PathNode* GetNext(void) const { return next; }; // return "next"
	void	SetNext(PathNode* n) { next = n; };	// set "next" to parm n
};	

//
//	Class Name : OnePath
//
//	Description : 
//		A list of PathNodes to represent the path from
//		one node (src_node) to another (dest node)
//
// 	Note: Start node (src_node) of any path is not stored in
// 	      node_list, but end node (dest_node) is.
//
//	Note: A better name for this class is simply "Path".
//
class OnePath {
private:
	PathNode* node_list;  // list of nodes on the path, excluding start node
			      // node_id of last PathNode is dest_node
	int	src_node;     // start node of the path
	int  	dest_node;    // end node of the path
	int	weight_sum;   // sum of ccosts of edges on the path
	int	delay_sum;    // sum of delays of edges on the path
	int	hops;	      // path length
	OnePath*	next; // to form a list of several paths
	int 	DestNodeReqBand; 
	   // if the path belongs to an MC tree (treePaths), this is the requested
	   // bandwidth of the destination node (participant) of the path.

public:
	OnePath() {              // init the attributes
		node_list = NIL;
		src_node = -1;
		dest_node = -1;
		weight_sum = 0;
		delay_sum = 0;
		hops = 0;
		next = NIL;
		DestNodeReqBand = 0;
	}

	OnePath(OnePath* data);

	OnePath(int s_node, int w, int d) {  // init the attributes
		node_list = NIL;
		src_node = s_node;
		dest_node = -1;    // end node not yet determined
		weight_sum = w;
		delay_sum = d;
		hops = 0;
		next = NIL;
		DestNodeReqBand = 0;
	}

	~OnePath();

	int 	GetDestNodeReqBand(void) const {return DestNodeReqBand;}
	void 	SetDestNodeReqBand(int dataBand) {DestNodeReqBand = dataBand;}		

	PathNode*  GetPathNode() const { return node_list; }
		   // return pointer to list of PathNodes
	void 	SetPathNode(PathNode* p) { node_list = p; }
		// set pointer to list of PathNodes to parm "p"
	void 	AddNodeToEnd(int n);
		// add new PathNode (with node ID "n") to end of node_list
		// not used for the time being
	void 	AddNodeToHead(int n);
		// add new PathNode (with node ID "n") to front of node_list

	int 	GetSrcNode(void) const { return src_node; };  // return src_node
	void 	SetSrcNode(int d) { src_node = d; } // set src_node to parm "d"

	int 	GetDestNode(void) const { return dest_node; }; // return dest_node
	void 	SetDestNode(int d) { dest_node = d; } // set dest_node to parm "d"

	int 	GetWeight(void) const { return weight_sum; }; // return weight_sum
	void 	SetWeight(int w) { weight_sum = w; }; // set weight_sum to parm "w"

	int 	GetDelay(void) const { return delay_sum; };  // return delay_sum
	void 	SetDelay(int d) { delay_sum = d; };  // set delay_sum to parm "d"

	int	GetHop(void) const { return hops; };  // return path length
	void	SetHop(int h) { hops = h; };  // set path length to parm "h"

	OnePath* GetNext(void) const { return next; }  // return "next"
	void 	SetNext(OnePath* n) { next = n; }  // set "next" to parm "n"

	COST	FindOldCost(Graph* graph);	// Forward path
	COST	FindNewCost(Graph* graph);	// Backward path
						// where is implementation?
	COST	FindCost(Graph* graph, PATH_TYPE p_type);
		// returns cost of the path (shared path or new path)
		// implementation in path.cc

	BOOL    IsInPath(int s_node);
		// returns TRUE if node with ID "s_node" is on the path.
		// returns FALSE otherwise.
		// implementation in path.cc

	void	GenBackPath(const OnePath* one_path);
		// reverse path "one_path" (switching start node and end node).
		// reversed path is stored in another object (other than "one_path").
		// implementation in path.cc

	BOOL	ConvertBackToFor(void);
		// reverse the path (switching start node and end node).
		// the new reversed path is stored in the same object.
		// implementation in path.cc

	void 	Display();  	   // display attributes on screen
	void 	Display(FILE* fp); // print attributes to a file
				   // implementation in path.cc

	void 	AddPathToEnd(OnePath* secondPath);  
		// add "secondPath" to the end of current path
	void 	SetCosts(COST dataCosts);
	int 	FindPreceedingNode(int thisNodeID);
};	  	

//
//	Class Name : Path
//
//	Description : 
//		A list of several paths (OnePath's) from a start node
//		to different end nodes.
//
//		Mainly used by ALG class and SESSION class
//
//	Note: A better name for this class is "Paths"
//
class Path {
private:
	int	src_node;  // Start node
	OnePath* m_path;   // List of paths from start node to different end nodes
			   // m_path is the head of the list
	ADJ_TYPE type;	   // Adjacency list type (FORWARD or BACKWARD)

public:
	Path() {		// init attributes
		src_node = -1;
		m_path = NIL;
		type = FORWARD;
	}

	Path(int n, ADJ_TYPE t = FORWARD) {	// init attributes
		src_node = n;
		m_path = NIL;
		type = t;
	}


	~Path();

	void	SetSrcNode(int n) { src_node = n; };  // set start node ID to parm "n"
	int	GetSrcNode(void) const { return src_node; }  // return start node ID

	void	AddOnePath(OnePath* path);
		// add a new path (parm "path") to the front of list of paths
		// implementation in path.cc

	OnePath* GetPath(void) const { return m_path; }  // return head of list of paths

	ADJ_TYPE GetAdjType(void) const { return type; }  // return FORWARD or BACKWARD
	void	 SetAdjType(ADJ_TYPE adj_type) { type = adj_type; }
		// set "type" to parm "adj_type"

	OnePath* FindOnePathEndWith(int s_node);
		 // returns the path whose end node ID is "s_node".
		 // returns NIL pointer if no such path is found.
		 // implementation in path.cc

	BOOL	ConvertBackToFor(void);
		// reverse every path in the list of paths
		// implementation in path.cc

	BOOL	AddWholePath(Path* new_path);
		// implementation in path.cc; not used for the time being

	void	Display();  // display attributes of every path in list of paths
			    // implementation in path.cc
	void 	Display(FILE *fp);

	OnePath* FindOnePathStartWith(int startNodeID);
		 // returns the path whose start node ID is "startNodeID".
		 // returns NIL pointer if no such path is found.
		 // implementation in path.cc

	BOOL 	RemovePath(int endNode);
		// remove path whose end node ID is "endNode".
		// returns FALSE if no such path is found.
		// return TRUE if the removal is successful.
};
#endif


///////////////////////////////////////////////////////////////////////////////////
//
// Updates
// Feb 09: added method FindOnePathStartWith (UT)
// Feb 16: added attribute destNodeRedBand to class OnePath (UT)
// Feb 17: added method FindPreceedingNode() to class OnePath (UT)
// Feb 24: added method RemovePath() to class Path (UT)
//
///////////////////////////////////////////////////////////////////////////////////