///////////////////////////////////////////////////////////////////////////////
//
//	File name : session.cc			Last updated: April 30, 98
//
//	Description :
//		This file implements methods of class Session.
//
//	Authors : Jin-Hwee Park / U.T.
// 	Date 	: Feb-01-1998
//
///////////////////////////////////////////////////////////////////////////////

#include "session.h"

//
// Initialize current Session object.
// Note: Method InitMCTree(graph)
// The list of NodeItems of "graph" is copied to the MC tree, but only
// node IDs are copied.  The other attributes of NodeItems of the
// MC tree are init to NIL or 0. "graph" is the network graph.
//
Session::Session(int id, int s_node, Graph* graph) {
	session_id = id;	// session ID
	sess_src_node = s_node; // mcSource node ID

	tot_node_num = graph->GetNodeNum();  // tot_node_num = graph->num_of_nodes
	in_session = new STATUS[tot_node_num];
		// dynamically alloc array, one element for each node of graph
	numDependants = new int[tot_node_num];
	for (int i=0; i<tot_node_num; i++) 
	{
 	  numDependants[i] = 0;
	  in_session[i] = UNSEEN;  // at beginning, no node of graph is in MC tree ...
	}
	SetNodeStatus(s_node, LEAF_NODE);  // ... except the mcSource

	avg_num_part = 0;  // no participants yet
	treePaths = new Path(s_node);  // alloc a new object; init src_node = s_node
	InitMCTree(graph);
	stat = new SessionStatistics;	 // alloc a new object
	next = NIL;
}

//
// Deallocate dynamic data structures
//
Session::~Session()
{
	delete [] in_session;
	delete [] numDependants;

	NodeItem* tmp;
	while(mc_tree != NIL) {
		tmp = mc_tree;
		mc_tree = mc_tree->GetNext();  // mc_tree = mc_tree->next;
		delete tmp;
	}

 	delete treePaths;
	delete stat;
}

//
// Initialize the MC tree.
// The list of NodeItems of "graph" is copied to the MC tree, but only
// node IDs are copied.  The other attributes of NodeItems of the
// MC tree are init to NIL or 0.
//
void	Session::InitMCTree(Graph* graph)
{
	NodeItem* tmp;

	tmp = graph->GetHead();  // tmp = head of list of nodes of "graph"
	mc_tree = NIL;

	while(tmp != NIL) {
		int  node_id = tmp->GetNode()->GetNodeID(); // tmp->node->node_id
		Node* new_node = new Node(node_id);  // alloc a new Node object
						     // init node ID to "node_id"
		NodeItem* new_nd_item = new NodeItem; // alloc new NodeItem object
		new_nd_item->SetNode(new_node); // new_nd_item->node = new_node;
		AddNodeToMC(new_nd_item);  // add new_nd_item to the end of
					   // list of NodeItems of MC tree
		tmp = tmp->GetNext();  // tmp = tmp->next;
	}
}

//
// Add NodeItem "nd_item" to the end of the list of NodeItems of the MC tree
//
void	Session::AddNodeToMC(NodeItem* nd_item)
{
	if (mc_tree == NIL) mc_tree = nd_item; // first and only item
	else {
		NodeItem* tmp;
		tmp = mc_tree;
		while(tmp->GetNext() != NIL)   // (tmp->next != NIL)
			tmp = tmp->GetNext();  // tmp = tmp->next;
		tmp->SetNext(nd_item);	       // tmp->next = nd_item;
	}
}

//
// Traverse the list of NodeItems of the MC tree and
// return pointer to the Node whose node ID is "node_id".
// Return NIL pointer if no such node is found.
//
Node*	Session::FindNode(int node_id)
{
	NodeItem* tmp = mc_tree;  // tmp = head of list of NodeItems of MC tree

	while(tmp != NIL) {  	  // for every NodeItem of MC tree
		Node* nd;

		nd = tmp->GetNode();  		 // nd = tmp->node;
		if (nd->GetNodeID() == node_id)  // (nd->node_id == code_id)
			return nd;
		tmp=tmp->GetNext();   		 // tmp = tmp->next;
	}
	return NIL;   // no such node is found
}

//
// Given a start node with ID = "node_id", return pointer to
// the adjacency edge whose end node ID is "dest_node".
// Return NIL if no such edge is found.
//
Edge*	Session::FindEdge(int node_id, int dest_node)
{
	Node* nd = FindNode(node_id); 	// get pointer to Node from node_id
	EdgeItem* edge_item;

	edge_item = nd->GetEdgeItem();  // edge_item = nd->for_adjlist;
	while(edge_item != NIL) {
		if (edge_item->GetEdge()->GetTargetNodeID() == dest_node)
		// (edge_item->edge->target_node_id == dest_node)
			return edge_item->GetEdge();  // edge_item->edge
		edge_item = edge_item->GetNext(); // edge_item = edge_item->next
	}
	return NIL;   // no such edge is found
}

//
// Display attributes of this session
//
void	Session::Display() 
{
	printf("* SESSION INFO :  session_id[%d], Source node[%d]\n",
		GetSessID(), GetSrcNode());
	for (int i=0; i<tot_node_num; i++) 
		if (in_session[i] == LEAF_NODE) 
			printf("LEAF_NODE[%d] ", i);  // participant node
		else if (in_session[i] == INTREE)
			printf("INTREE[%d] ", i);     // intermediate node
	printf("\n");
}


//////////////////////////////////////////////////////////////////////////////
//
//	Method name : FindPathFromSrcToDest
//
// 	Description :
// 		Find the path from mcSource node to every other
//		node in the multicast tree.
//
//	Return : Path* T
//		The list of paths, each path connecting mcSource
//		and another node in the multicast tree.
//
//
//	Note : the path from mcSource to a node in the tree is unique.
//
//	Input : MC tree, represented by adjacency lists, an adj list
//		for each node in the tree.
//
//	Output: MC tree, represented by paths, a path from mcSource
//		to each node of the tree.
//
//////////////////////////////////////////////////////////////////////////////
Path*	Session::FindPathFromSrcToDest()
{
	int	tot_nodes = GetNodeNum();
	int	source_node = GetSrcNode();   // get mcSource node ID
	int	*parent = new int[tot_nodes];

	NodeItem*	nd_item_head;
	EdgeItem*	edge_item;

	for (int i = 0; i<tot_nodes; i++) parent[i] = -1;  // init parent[]

	nd_item_head = GetMCTree();   // nd_item_head = mc_tree
				      // (= head of list of NodeItems)
	while(nd_item_head != NIL) {  // for every NodeItem in MC tree list
		int	nd_id = nd_item_head->GetNode()->GetNodeID();
		     // nd_id = nd_item_head->node->node_id;
		int	index;

		edge_item = nd_item_head->GetNode()->GetEdgeItem();
		  // edge_item = nd_item_head->node->for_adjlist;
		while(edge_item != NIL) {  // for every edge in adj list
			index = edge_item->GetEdge()->GetTargetNodeID();
			   // index = edge_item->edge->target_node_id;
			   // nd_id is start node; index is end node
			parent[index] = nd_id;  // parent means start node (of "index")
			edge_item = edge_item->GetNext();
			   // edge_item = edge_item->next;
		}

		nd_item_head = nd_item_head->GetNext();
		   // nd_item_head = nd_item_head->next;
	}  // end while(nd_item_head != NIL)

	Path* path = new Path(source_node, FORWARD);
	   // path will store a list of paths from source_node
	   // to the other nodes in the MC tree

	for (int child_node=0; child_node<tot_nodes; child_node++) {
		int	weight, delay;
		if (child_node == source_node) continue;

		// if a node does not have a parent, do not generate path
		// since the node does not belong to the MC tree.
		if (parent[child_node] == -1) continue;

		STATUS status = GetNodeStatus(child_node);

		if (status == LEAF_NODE || status == INTREE) {
			Edge* edge = FindEdge(parent[child_node], child_node);
			if (edge == NIL) {
				printf("FILE[%s], LINE[%d] : fatal error !\n",
					__FILE__, __LINE__);
				Display_tree();
				printf("Edge [%d ->%d]\n", parent[child_node], child_node);
				exit(1);
			}
			weight = edge->GetWeight();
			delay = edge->GetDelay();
			//
			// This edge stores costs of the path from mcSource to 
			// child_node.  See alg_util.cc, method UpdateTreeWithPath.
			// That's why AddNodeToHead need not accumulate the
			// weights/delays of edges on the path.
			//
		} else continue;

		OnePath* one_path = new OnePath(source_node, weight, delay);
			// alloc a new path; init the above attributes.
			// "weight", "delay" are costs of the PATH from
			// source_node to child_node.
		one_path->AddNodeToHead(child_node);
			// add child_node to the front of one_path
		//
		// Trace back the path from child_node to mcSource node
		// and add each encountered edge to the front of one_path
		//
		int cur_node = parent[child_node];
		while (cur_node != source_node) {
			one_path->AddNodeToHead(cur_node);
				// add cur_node to the front of one_path
			cur_node = parent[cur_node];  // trace back one edge
		}
		path->AddOnePath(one_path);  // add one_path to list of paths
	}

	delete [] parent;
	return path;  // return list of paths from mc Source to the other nodes
}

//
// If the new path (segment) crosses a node belonging to the MC tree,
// return TRUE.  Return FALSE otherwise.
// Note:
// Parameter one_path should be a new path (segment) returned by
// Dijkstra's algorithm.  The start node of the path is a node
// in the connection set (and thus belonging to the tree).
//
// (The term "cycle" should be replaced by "loop")
//
BOOL    Session::IsCycle(OnePath* one_path)
{
#ifdef FUNCTION_TRACE
  fprintf(stderr, "++++++++++++++ IsCycle() (cnode [%d])\n", one_path->GetSrcNode());
#endif

	PathNode* path_node = one_path->GetPathNode();  // one_path->node_list

	if (path_node == NIL) 
	{
	  printf("NIL path in IsCycle()\n");
	  return TRUE;
	}

	while(path_node != NIL) {
		int cur_node = path_node->GetNodeID();  // path_node->node_id

		if (GetNodeStatus(cur_node) != UNSEEN)  // cur_node is in tree
		{
#ifdef DEBUG_LEVEL_1
		  printf("New path contains cycle (connection node [%d]): \n",
						one_path->GetSrcNode());
		  one_path->Display();
#endif
		  return TRUE;
		}
		path_node = path_node->GetNext();
	}
	return FALSE;
}

//
// Return TRUE if the node with node_id in in the MC tree.
// Return FALSE otherwise.
//
BOOL	Session::IsNodeInTree(int node_id)
{
	NodeItem* nd_item = GetMCTree();  // nd_item = mc_tree
					  // (head of list of NodeItems)
	while(nd_item != NIL) {
		EdgeItem* edge_item;

		edge_item = nd_item->GetNode()->GetEdgeItem();
			//  nd_item->node->for_adjlist;
		while(edge_item != NIL) {
			if (edge_item->GetEdge()->GetTargetNodeID() == node_id)
			 // edge_item->edge->target_node_id
				return TRUE;
			edge_item = edge_item->GetNext();  // edge_item->next
		}
		nd_item = nd_item->GetNext();  // nd_item->next
	}
	return FALSE;

// Note: why don't we check in_session[node_id] ???
}

// Copy attributes "total_band" and "current_band" from parm "edge"
// to the edge in the MC tree whose start node ID is "src_node".
// Parameter "edge" is an Edge object of the network graph.
//
void	Session::UpdateEdgeData(const int src_node, const Edge* edge)
{
	int	t_node = edge->GetTargetNodeID();  // edge->target_node_id

	if (GetNodeStatus(t_node) == UNSEEN) return;  // node not in tree

	Edge*	edge_in_sess = FindEdge(src_node, t_node);
		// find edge that connects "src_node" and "t_node".
		// "edge_in_sess" is an Edge object of the MC tree.

	if (edge_in_sess == NIL) {
		printf("FILE[%s], LINE[%d] : invalid edge pointer !!\n",
			__FILE__, __LINE__);
		printf("Edge [%d -> %d]\n", src_node, t_node);
		exit(1);
	}
	edge_in_sess->UpdateEdge(edge);
		// copy attributes "total_band" and "current_band"
		// from "edge" to "edge_in_sess"
	return;
}

//
// Re-initialize the current Session object
// Note: A better name for this method is ReInitDynamicData
//
void	Session::DeleteDynamicData(Graph* graph)
{
	NodeItem* mc_nd_item = mc_tree;  // head of list of NodeItems
	while(mc_nd_item != NIL) {   // delete current MC tree
		NodeItem* tmp = mc_nd_item;
		mc_nd_item = mc_nd_item->GetNext();  // mc_nd_item->next
		delete tmp;
	}
	InitMCTree(graph);  // re-init the MC tree

	delete stat;
	stat = new SessionStatistics;

	for (int i=0; i<tot_node_num; i++) 
	{
	  numDependants[i] = 0;
	  in_session[i]    = UNSEEN;
	}
	SetNodeStatus(sess_src_node, LEAF_NODE);  // mcSource is a participant
}

//
// Display the list of NodeItems of the MC tree (outputs printed on screen)
// See comments in the next function (method).
//
void	Session::Display_tree()
{
	NodeItem* node_item;
	printf("\n*** MULTICAST TREE ===> Session[%d], MC_Source[%d]\n",
		GetSessID(), GetSrcNode());

	for (node_item  = mc_tree;
	     node_item != NIL;
	     node_item  = node_item->GetNext())
	{
	   Node* node   = node_item->GetNode();
	   int   nodeID = node->GetNodeID();		
	   if(in_session[nodeID] != UNSEEN)  // node is in the MC tree
	   {
		printf("N[%d] OE[%d] DP[%d] => ", nodeID,
			node->GetEdgeNum(FORWARD), numDependants[nodeID]);

		EdgeItem* edge_item = node->GetEdgeItem();
		while(edge_item != NIL) {
			Edge* edge;

			edge = edge_item->GetEdge();
			if (edge->GetEdgeStatus() == LEAF_NODE) {
				printf("*N[%d] W[%d] D[%d] B[%d] -> ",
					edge->GetTargetNodeID(),
					edge->GetWeight(), edge->GetDelay(), edge->GetCurBand());
			} else if (edge->GetEdgeStatus() == INTREE) {
				printf("&N[%d] W[%d] D[%d] B[%d] -> ",
					edge->GetTargetNodeID(),
					edge->GetWeight(), edge->GetDelay(), edge->GetCurBand());
			} else {
				printf("N[%d] W[%d] D[%d] B[%d] -> ",
					edge->GetTargetNodeID(),
					edge->GetWeight(), edge->GetDelay(), edge->GetCurBand());
			}
			edge_item = edge_item->GetNext();
		}  // end while(edge_item)
		printf("*\n");
	   }  // end if(in_session[nodeID])
	}  // end for(node_item)
}

//
// Display the list of NodeItems of the MC tree (outputs printed to a file)
// For every NodeItem, print the node ID and its adjacency list
// of edges, each edge with the weight, delay and available bandwidth.
//
void	Session::Display_tree(FILE* fp)
{
	NodeItem* node_item;

	node_item = mc_tree;  // node_item = head of list of NodeItems of MC tree

	fprintf(fp, "\n*** MULTICAST TREE ===> Session[%d], MC_Source[%d]\n",
		GetSessID(), GetSrcNode());  // print session ID and mcSource ID
	while(node_item != NIL) {  // for every NodeItem in MC tree
		EdgeItem* edge_item;

		fprintf(fp, "\tN[%d] => ", node_item->GetNode()->GetNodeID());
		   // print node ID (node_item->node->node_id)
		edge_item = node_item->GetNode()->GetEdgeItem();
		   // edge_item = node_item->node->for_adjlist;

		while(edge_item != NIL) {  // for every adjacency edge
			Edge* edge;

			edge = edge_item->GetEdge();  // edge = edge_item->edge;
			if (edge->GetEdgeStatus() == LEAF_NODE) {
				fprintf(fp, "*N[%d] W[%d] D[%d] B[%d] -> ",
					edge->GetTargetNodeID(),
					edge->GetWeight(), edge->GetDelay(), edge->GetCurBand());
			} else if (edge->GetEdgeStatus() == INTREE) {
				fprintf(fp, "&N[%d] W[%d] D[%d] B[%d] -> ",
					edge->GetTargetNodeID(),
					edge->GetWeight(), edge->GetDelay(), edge->GetCurBand());
			} else {  // UNSEEN (not in tree)
				fprintf(fp, "N[%d] W[%d] D[%d] B[%d] -> ",
					edge->GetTargetNodeID(),
					edge->GetWeight(), edge->GetDelay(), edge->GetCurBand());
			}
			edge_item = edge_item->GetNext();  // edge_item = edge_item->next;
		}
		fprintf(fp, "*\n");
		node_item = node_item->GetNext();  // node_item = node_item->next;
	}
}

//
// Display the paths from mcSource to the other nodes in the MC tree
//
void	Session::Display_path() 
{
	Path*	allpaths = FindPathFromSrcToDest();  // get all the paths

	printf("\n*** ALL PATHS ===> Session[%d], MC_Source[%d]\n", 
				GetSessID(), GetSrcNode());
	if (allpaths != NIL) {
		allpaths->Display();
		delete allpaths;
	}
}


void Session::DisplayTreePaths()
{
  printf("\n*** S2D PATHS ===> Session[%d], MC_Source[%d]\n", 
				GetSessID(), GetSrcNode());
  if (treePaths != NIL)
    treePaths->Display();
}


//
// For every node on the path (except the last node), increment 
// (or decrement) numDependants[] of the node.
//
void Session::UpdateNumDependants(OnePath* thePath, int numDep)
{
  PathNode* tmp;
  numDependants[thePath->GetSrcNode()] += numDep;
  for (tmp = thePath->GetPathNode();
       tmp->GetNext()!= NIL;
       tmp = tmp->GetNext())      
    numDependants[tmp->GetNodeID()] += numDep;
}


void Session::UpdateStatistics(Graph* graph, UserReq* userRequest, 
					BOOL requestStatus)
{
  stat->SetNumGraphNodes(graph->GetNodeNum());
  switch(userRequest->GetUserReq())
  {
    case JOIN:
      stat->IncNumJoins();
      if (requestStatus ==  FALSE_2)	// rejected
      {
	stat->IncNumRejections();
	return;		// no update needed
      }
      stat->IncSumReqPathLength(userRequest->GetQos().GetHop());
      stat->IncSumReqPathDelay(userRequest->GetQos().GetDelay());
      stat->IncSumReqBandwidth(userRequest->GetQos().GetBandwidth());
      break;

    case LEAVE:
      stat->IncNumLeaves();
      break;

    default:
      fprintf(stderr, "Invalid request type!\n");
      return;
  }

  // Update path length and delay statistics
  int numPaths 	= 0;
  int minPL 	= MAXINT;
  int maxPL 	= -1;
  int sumPL 	= 0;
  int minDelay 	= MAXINT;
  int maxDelay 	= -1;
  int sumDelay 	= 0;
  int pathLength;
  int pathDelay;

  OnePath* path;
  for(path = treePaths->GetPath(); path != NIL; path = path->GetNext())
  {
    numPaths++;
    pathLength = path->GetHop();
    if (minPL > pathLength)  minPL = pathLength;
    if (maxPL < pathLength)  maxPL = pathLength;
    sumPL += pathLength;   

    pathDelay  = path->GetDelay();
    if (minDelay > pathDelay) minDelay = pathDelay;
    if (maxDelay < pathDelay) maxDelay = pathDelay;
    sumDelay += pathDelay;
  }  // end for(path)

  if (numPaths > 0)
  {
    stat->SetMinPathLength(minPL);
    stat->SetMaxPathLength(maxPL);
    stat->SetMinPathDelay(minDelay);
    stat->SetMaxPathDelay(maxDelay);
    stat->SetAvgPathLength(sumPL / (float)numPaths);
    stat->SetAvgPathDelay(sumDelay / (float)numPaths);
  }
  else
  {
    stat->SetMinPathLength(0);
    stat->SetMaxPathLength(0);
    stat->SetMinPathDelay(0);
    stat->SetMaxPathDelay(0);
    stat->SetAvgPathLength(0.0);
    stat->SetAvgPathDelay(0.0);
  }

  // Update other statistics
  int totalNumNodes	= 0;
  int totalNumDests	= 0;
  int totalNumEdges	= 0;
  int maxNumOutEdges 	= -1;
  int maxNumDependants	= -1;
  int sumNumDependants	= 0;
  int numEdges		= 0;
  int usedBand		= 0;
  float normUsedBand	= 0.0;
  float mcTreeCost	= 0.0;
  float maxLinkCong	= -1.0;
  Edge* mostCongestedLink = NIL;
  float linkCong;
  int   sessionBand;			// bandwidth used by this session on an edge
  int   sessionID = GetSessID();	// session ID
  int 	mcSource  = GetSrcNode();	// mcSource of the session
  int   numDeps;

  NodeItem* nodeItem;
  EdgeItem* edgeItem;
  Node*	    node;
  Edge*	    grpEdge;
  int       nodeID;

  for(nodeItem = mc_tree; 
      nodeItem != NIL; 
      nodeItem = nodeItem->GetNext())
  {
    node   = nodeItem->GetNode();
    nodeID = node->GetNodeID();
    if (GetNodeStatus(nodeID) != UNSEEN)	// node is in the MC tree
    {
      totalNumNodes++;
      if (GetNodeStatus(nodeID) == LEAF_NODE) totalNumDests++;
      numEdges = node->GetEdgeNum(FORWARD);
      totalNumEdges += numEdges;
      if (maxNumOutEdges < numEdges)  maxNumOutEdges = numEdges;

      if (nodeID != mcSource)
		// Note: we should not consider the mcSource for this measure 
		// since all the destinations depend on the mcSource.
      {
      numDeps = GetNumDependants(nodeID);  // numDependants[nodeID]
      	if(maxNumDependants < numDeps)  
	  maxNumDependants = numDeps;
      	sumNumDependants += numDeps;
      }

      for(edgeItem = node->GetEdgeItem(); 
          edgeItem != NIL;
	  edgeItem = edgeItem->GetNext())
      {
	// Get the edge object from network graph to have bandwidth info
	grpEdge = graph->FindEdge(nodeID, edgeItem->GetEdge()->GetTargetNodeID());
	if (grpEdge == NIL)
	{
	  fprintf(stderr, "Fatal Error in UpdateStatistics(): NIL edge!\n");
	  exit(1);
	}
	sessionBand   = grpEdge->GetSessionUsedBandwidth(sessionID);
	usedBand     += sessionBand;
	normUsedBand += (float)sessionBand / (float)grpEdge->GetTotBand();
	mcTreeCost   += grpEdge->CongestionInc(grpEdge->GetEdgeUsedBand() - sessionBand,
						sessionBand);	// see Note (1)
	linkCong      = grpEdge->Congestion(grpEdge->GetEdgeUsedBand());
	if (maxLinkCong < linkCong)  
	{
	  maxLinkCong = linkCong;
	  mostCongestedLink = grpEdge;
	}
      }  // end for(edgeItem)

      stat->SetTotalNumNodes(totalNumNodes);
      stat->SetTotalNumDestinations(totalNumDests);
      stat->SetTotalNumEdges(totalNumEdges);
      stat->SetMaxNumOutEdges(maxNumOutEdges);
      stat->SetTotalUsedBandwidth(usedBand);
      stat->SetTotalNormUsedBandwidth(normUsedBand);
      stat->SetTreeCost(mcTreeCost);
      if (numPaths > 0)
      {
	stat->SetMaxNumDependants(maxNumDependants);
	stat->SetAvgNumDependants((float)sumNumDependants / (float)(totalNumNodes-1));
							// -1 to exclude the mcSource
      	stat->SetMaxLinkCongestion(maxLinkCong);
      }
      else
      {
	stat->SetMaxNumDependants(0);
	stat->SetAvgNumDependants(0.0);
      	stat->SetMaxLinkCongestion(0.0);
      }
      stat->SetMostCongestedLink(mostCongestedLink);

    }  // end if(GetNodeStatus)
  }  // end for(nodeItem)
}  // end UpdateStatistics

// Note: 
// (1) CongestionInc(BW used by the other sessions, BW used by this session)


void SessionStatistics::Display()
{
  printf("Statistics \n");
  printf("---------- \n");
  printf("Joins \t\t%d \tLeaves \t\t%d \tRejections \t%d \n", 
		numJoins, numLeaves, numRejections);
  printf("NumTreeNodes \t%d \tNumDests \t%d \tNumEdges \t%d \n",
		totalNumNodes, totalNumDestinations, totalNumEdges);
  printf("MinPathLength \t%d \tMaxPathLength \t%d \tAvgPathLength \t%.4f \n",
		minPathLength, maxPathLength, avgPathLength);
  printf("MinDelay \t%d \tMaxDelay \t%d \tAvgDelay \t%.4f \n",
		minPathDelay, maxPathDelay, avgPathDelay);
  printf("MaxOutEdges \t%d \tMaxDependants \t%d \tAvgDependants \t%.4f \n",
		maxNumOutEdges, maxNumDependants, avgNumDependants);
  printf("TreeCost \t%.4f \tTotalBandwidth \t%d \tTotalNormBw \t%.4f \n",
		treeCost, totalUsedBandwidth, totalNormUsedBandwidth);
  printf("MaxLinkCong \t%.4f ", maxLinkCongestion); 
  if (mostCongestedLink != NIL)
    printf("\tMostCongLink \t[%d->%d]", mostCongestedLink->GetSrcNode()->GetNodeID(), 
		mostCongestedLink->GetTargetNodeID());
  printf("\n");

  // Print statistics required by each specific experiment
  printf("EXP1 TreeCost %.8f AvgDelay %.2f AvgLength %.2f NumDests %d Rejections %d Joins %d Leaves %d \n", 
		treeCost, avgPathDelay, avgPathLength, totalNumDestinations, 
		numRejections, numJoins, numLeaves);
}


void SessionStatistics::DisplayFinalStat()
{
  printf("\nFinal Statistics\n");
  printf("----------------\n");
  printf("FINAL GraphNodes \t%d \tNumTreeNodes \t%d \tNumDests \t%d \n",
		numGraphNodes, totalNumNodes, totalNumDestinations);  
  printf("FINAL AvgReqPLength %.4f - AvgReqDelay %.4f - AvgReqBwidth %.4f \n",
		(float)sumReqPathLength / (float)numJoins, 
		(float)sumReqPathDelay  / (float)numJoins,
		(float)sumReqBandwidth  / (float)numJoins);
  printf("FINAL AvgPathLength %.4f - AvgDelay %.4f - TotalBandwidth %d - Rejections %d \n",
		avgPathLength, avgPathDelay, totalUsedBandwidth, numRejections);
  printf("EXP2 TreeCost %.8f AvgDelay %.2f AvgLength %.2f NumDests %d Rejections %d Joins %d Leaves %d AvgTreeCost %.10f \n", 
		treeCost, avgPathDelay, avgPathLength, totalNumDestinations, 
		numRejections, numJoins, numLeaves, treeCost/(float)totalNumDestinations);
}


//
// For the Shortest Path algorithm
// Return TRUE if there is no loop, and FALSE otherwise.
// Separate wholePath into a sharedPath and newPath (if a sharedPath exists).
//
BOOL Session::IsCycle_SP(OnePath* wholePath, OnePath* sharedPath, OnePath* newPath)
{
#ifdef FUNCTION_TRACE
  fprintf(stderr, "++++++++++++++ IsCycle_SP() \n");
  if (wholePath == NIL) fprintf(stderr, "---------------- NIL whole path \n");
#endif

  PathNode* pathNode = wholePath->GetPathNode();  // wholePath->node_list
  if (pathNode == NIL)
  {
    printf("NIL path in IsCycle_SP()\n");
    return TRUE;	// error condition
  }

  int endNode;
  int srcNode = wholePath->GetSrcNode();	// mcSource node
  sharedPath->SetSrcNode(srcNode);
  sharedPath->SetDestNode(srcNode);
  int startNode = srcNode;
  BOOL loopExists = FALSE;
  BOOL shPath = TRUE;
  for( ; pathNode != NIL; pathNode = pathNode->GetNext(), startNode = endNode)
  {
    endNode = pathNode->GetNodeID(); 
    Edge* curEdge = FindEdge(startNode, endNode);

    if (shPath == TRUE && curEdge != NIL)  // edge is in MC tree
      sharedPath->AddNodeToEnd(endNode);

    else if (shPath == TRUE && curEdge == NIL)
    {
      if (GetNodeStatus(endNode) == UNSEEN)  // endNode not in MC tree
      {
      	shPath = FALSE;		// get out of the sharedPath, begin the newPath
      	newPath->SetSrcNode(sharedPath->GetDestNode());
	   // start node of newPath is the last node of sharedPath (connection node)
      	newPath->AddNodeToEnd(endNode);
      }
      else  // both nodes in MC tree but connected by another edge => contains loop
	loopExists = TRUE;
    }

    else if (shPath == FALSE && curEdge == NIL)
    { 
      if (GetNodeStatus(endNode) == UNSEEN)	// then still on the new path
        newPath->AddNodeToEnd(endNode);
      else	// endNode in MC tree => get back to the tree => contains loop
	loopExists = TRUE;
    }

    else if (shPath == FALSE && curEdge != NIL)
      loopExists = TRUE;
  }  // end for

  if (loopExists == TRUE)
  {
#ifdef DEBUG_LEVEL_1
    printf("Whole path contains cycle: \n");
    wholePath->Display();
#endif
    return TRUE;	// loop exists
  }

  if (shPath == TRUE)	// the whole path is in MC tree => impossible
  {
    fprintf(stderr, "IsCycle_SP(): should not get here!\n");
    return TRUE;	// error condition
  }

  return FALSE;		// no loop

  // Note: the costs of sharedPath and newPath will be computed and 
  //       stored by function FindCandidateNewPaths_SP().
}  // end IsCycle_SP

/////////////////////////////////////////////////////////////////////////////////////
//
// Updates
// Feb 12: added attribute treePaths to class Session (UT)
// Feb 17: added method DisplayTreePaths() (UT)
// Feb 19: added "printf" to IsCycle() (UT)
//
/////////////////////////////////////////////////////////////////////////////////////