// session.h					Last updated: May 15, 98

#ifndef Session_H
#define Session_H

#include "common.h"
#include "path.h"


///////////////////////////////////////////////////////////////////////////////
//
//	Class Name : SessionStat
//
//	Description : 
//		Class to contain the statistic information for a session
//
///////////////////////////////////////////////////////////////////////////////
class SessionStatistics
{
private:
  int	numJoins;		// total number of joins
  int	numLeaves;		// total number of leaves
  int 	numRejections;		// number of joins being rejected 
				// Note: leaves always succeed.
  // QoS
  int sumReqPathLength;		// sum of requested path lengths of join requests
  int sumReqPathDelay;		// sum of requested end-to-end delays of join requests
  int sumReqBandwidth;		// sum of requested bandwidths of join reques

  // End-to-end performance
  int	minPathLength;		// S2D path length
  int 	maxPathLength;
  float	avgPathLength;

  int 	minPathDelay;		// end-to-end delay
  int 	maxPathDelay;
  float	avgPathDelay;

  // Link utilization
  				// number of outgoing edges from a node
  //int	minNumOutEdges;		// always 0	
  int 	maxNumOutEdges;
  //float avgNumOutEdges;	// always 1 (approx.) for a tree

  				// number of destinations depending on a node
  //int minNumDependants;	// always 0	
  int 	maxNumDependants;
  float avgNumDependants;

  // Efficiency of the tree
  int	totalNumEdges;		// total number of edges in the MC tree

  float maxLinkCongestion;	// max link congestion
  Edge* mostCongestedLink;	// the most congested link

  int	totalUsedBandwidth;	// total bandwidth allocated to this session
  float totalNormUsedBandwidth;	// total normalized bandwidth allocated to this session
				// sum of (sessionUsedBandwidth / edgeTotalBandwidth)
  float treeCost;		// sum of (congestion_increase(edge e)) 
				// for every edge e in the MC tree

  // Others
  int	totalNumNodes;		// total number of nodes in the MC tree
  int	totalNumDestinations;	// total number of destination nodes in the MC tree
  int 	numGraphNodes;		// total number of nodes in the network graph

public:
  SessionStatistics()
  {
	numJoins 		= 0;
	numLeaves		= 0;
	numRejections		= 0;
   	sumReqPathLength	= 0;
  	sumReqPathDelay		= 0;
  	sumReqBandwidth		= 0;
	minPathLength		= MAXINT;
	maxPathLength		= 0;
	avgPathLength		= 0.0;
	minPathDelay		= MAXINT;
	maxPathDelay		= 0;
	avgPathDelay		= 0.0;
	//minNumOutEdges	= 0;		// always 0 (leaf nodes)
	maxNumOutEdges		= 0;
	//avgNumOutEdges	= 0.0;		// always 1 (approx.) for a tree
	//minNumDependants	= 0;		// always 0 (leaf nodes)
	maxNumDependants	= 0;
	avgNumDependants	= 0.0;
	numGraphNodes		= 0;
	totalNumNodes		= 0;
	totalNumDestinations	= 0;
	totalNumEdges		= 0;
	maxLinkCongestion	= 0.0;
	mostCongestedLink	= NIL;
	totalUsedBandwidth	= 0;
	totalNormUsedBandwidth	= 0.0;
	treeCost		= 0.0;
  }

  ~SessionStatistics() {return;}
  void IncNumJoins() 		{numJoins++;}
  void IncNumLeaves()		{numLeaves++;}
  void IncNumRejections()	{numRejections++;}

  void IncSumReqPathLength(int newData) {sumReqPathLength += newData;}
  void IncSumReqPathDelay(int newData) 	{sumReqPathDelay += newData;}
  void IncSumReqBandwidth(int newData) 	{sumReqBandwidth += newData;}

  void SetMinPathLength(int data)	{minPathLength = data;}
  void SetMaxPathLength(int data)	{maxPathLength = data;}
  void SetAvgPathLength(float data)	{avgPathLength = data;}

  void SetMinPathDelay(int data)	{minPathDelay = data;}
  void SetMaxPathDelay(int data)	{maxPathDelay = data;}
  void SetAvgPathDelay(float data)	{avgPathDelay = data;}

  void SetMaxNumOutEdges(int data)	{maxNumOutEdges = data;}
  void SetMaxNumDependants(int data)	{maxNumDependants = data;}
  void SetAvgNumDependants(float data)	{avgNumDependants = data;}

  void SetNumGraphNodes(int data)		{numGraphNodes = data;}
  void SetTotalNumNodes(int data)		{totalNumNodes = data;}
  void SetTotalNumDestinations(int data)	{totalNumDestinations = data;}
  void SetTotalNumEdges(int data)		{totalNumEdges = data;}

  void SetTotalUsedBandwidth(int data)		{totalUsedBandwidth = data;}
  void SetTotalNormUsedBandwidth(float data)	{totalNormUsedBandwidth = data;}
  void SetTreeCost(float data)			{treeCost = data;}
  void SetMaxLinkCongestion(float data)		{maxLinkCongestion = data;}
  void SetMostCongestedLink(Edge* edge)		{mostCongestedLink = edge;}

  void Display();
  void DisplayFinalStat();
};

///////////////////////////////////////////////////////////////////////////////
//
//	Class Name : Session
//
//	Description :
//		This class is representation of a session in the network.
//		Each session has only dynamic data generated by Algorithm class
//		with a set of group requests -- a multicasting tree and statistic data
//
//		Every session is represented by an object of class Session.
// 		Several sessions are linked into a linked list.
//
// 	Note:
//		"in_session" is a dynamically allocated array, one element for
//		each node of the graph.  Each element is indexed by a node ID
//		and indicates the status of that node in the MC tree.
//
///////////////////////////////////////////////////////////////////////////////
class Session {

   private:
	int	session_id;	// Session identifier
	int	tot_node_num;	// Total number of nodes in the graph
	int	sess_src_node;	// ID of mcSource node
	STATUS 	*in_session;	// List of nodes in the session
				// INVALID, LEAF_NODE
	int	avg_num_part;	// Average number of participants
	NodeItem *mc_tree;	// Resulting multicasting tree
	SessionStatistics *stat;// Session statistic
	Session* next;		// Pointer to next session in list of sessions
	Path* treePaths;	// list of paths from mcSource to participant nodes
	  // Note: to find the list of paths from mcSource to ALL the nodes 
	  // of the MC tree (including intermediate nodes), invoke method 
	  // Session::FindPathFromSrcToDest().
	int 	*numDependants; // array of integers, one element per node
				// numDependants[i]: number of destinations 
				// depending on node i

   public:
      	Session(int id, int s_node, Graph* graph);
      	~Session();

	Path* GetTreePaths(void) const {return treePaths;}
	void  SetTreePaths(Path* listPaths) {treePaths = listPaths;}

	int	GetSessID(void) const { return session_id; }
		// return session ID

	int	GetNodeNum(void) const { return tot_node_num; }
		// return total number of nodes of the graph
	void	SetNodeNum(int n) { tot_node_num = n; }
		// set total number of nodes of the graph to "n"
	void	IncNodeNum(void) { tot_node_num++; }
		// increment total number of nodes of the graph

	int	GetSrcNode(void) const { return sess_src_node; }
		// return mcSource node ID
	void	SetSrcNode(int n) { sess_src_node = n; }
		// set mcSource node ID to parm "n"

	STATUS	GetNodeStatus(int i) { return in_session[i]; }
		// return status of node with ID "i"
	void	SetNodeStatus(int index, STATUS s) { in_session[index] = s; }
		// set status of node with ID "index" to parm "s"

	NodeItem* GetMCTree(void) const { return mc_tree; }
		  // return pointer to head of list of NodeItems of MC tree
	void 	InitMCTree(Graph* graph);
		// List of NodeItems of "graph" is copied to MC tree, but
		// only node IDs are copied.  The other attributes of
		// NodeItems of the MC tree are init to NIL or 0.
	void	AddNodeToMC(NodeItem* nd_item);
		// add "nd_item" to the end of list of NodeItems of MC tree
	
	SessionStatistics* GetSessionStat(void) const { return stat; }  // return "stat"

	Session* GetNext(void) const { return next; } // return "next"
	void 	SetNext(Session* n) { next = n; }  // set "next" to parm "n"

	Node*	FindNode(int node_id);
		// traverse the list of NodeItems of the MC tree and
		// return pointer to the Node whose node ID is "node_id".
		// return NIL pointer if no such node is found.
	Edge*   FindEdge(int node_id, int dest_node);
		// given a start node with ID = "node_id", return pointer to
		// the adjacency edge whose end node ID is "dest_node".
		// return NIL if no such edge is found.

	// Find every path from source node to every other node in the MC tree
        Path*   FindPathFromSrcToDest();

	// Check whether the path contains a cycle
	BOOL 	IsCycle(OnePath* one_path);
	BOOL	IsCycle_SP(OnePath* wholePath, OnePath* sharedPath, OnePath* newPath);

	// Check the status of nodes in the multicasting tree
	BOOL    IsNodeInTree(int node_id);

	// Update bandwidth information of edge with given edge information
	// in the multicasting tree
	void	UpdateEdgeData(const int src_node, const Edge* edge);

	// Initialize the session dynamic data to execute multicasting
	// algorithm
	void	DeleteDynamicData(Graph* graph);

	void	Display();  // display attributes of this session
	void	Display_tree();
	void	Display_tree(FILE* fp);
	void	Display_path();
	void	DisplayTreePaths();

	BOOL 	InConnectionSet(int node_id) 
		{ if (GetNodeStatus(node_id) == UNSEEN) return FALSE; return TRUE; }
	void	UpdateStatistics(Graph *graph, UserReq* userRequest, 
					BOOL requestStatus);

	int 	GetNumDependants(int nodeID)	{return numDependants[nodeID];}
	void	SetNumDependants(int nodeID, int numDep) 
						{numDependants[nodeID] = numDep;}
	void 	UpdateNumDependants(OnePath* thePath, int numDep);
};

#endif


/////////////////////////////////////////////////////////////////////////////////////
//
// Updates
// Feb 06: added method InConnectionSet (UT)
// Feb 12: added attribute treePaths to class Session (UT)
// Feb 17: added method DisplayTreePaths() (UT)
//
/////////////////////////////////////////////////////////////////////////////////////